package game;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.TreeSet;

import util.Direction;

public class AITools {
  private static final int BEAM_WIDTH = 50;
  private static final int MAX_BEAM_DEPTH = 10;
  private static final int MINIMAX_BRANCH = 5;
  
  public static State getNextMove(Bullies game, int minimaxDepth) {
    Node root = new Node(game);
    Node result = performMinimax(root, game, game.getActivePlayer(), minimaxDepth);
    Node nextMove = result;
    while (nextMove.parent.parent != null) {
      nextMove = nextMove.parent;
    }
    nextMove.state.setCallNext(false);
    System.out.println("------------------------------");
    
    return nextMove.state;
  }
  
  // Recursively looks for best state via minimax
  public static Node performMinimax(Node parent, Bullies game, Player evaluator, int depth) {
    if (depth < 1) {
      System.out.println("Leaf: " + parent.heuristic);
      parent.alphaBetaMax = parent.heuristic;
      parent.alphaBetaMin = parent.heuristic;
      return parent;
    } else {
      State.pop(game, parent.state);
      boolean isMax = game.getActivePlayer().equals(evaluator);
      System.out.println("Checking depth " + depth + " (" + game.getActivePlayer().getName() + ")");
      
      Node best = null;
      TreeSet <Node> openSet = evaluateTurn(game, BEAM_WIDTH, isMax, evaluator);
      Iterator <Node> itr;
      if (isMax) itr = openSet.descendingIterator();
      else itr = openSet.iterator();
      
      for (int i=0; i < Math.min(openSet.size(), MINIMAX_BRANCH); ++i) {
        Node child = itr.next();
        child.parent = parent;
        child.state.setCallNext(true);
        child = performMinimax(child, game, evaluator, depth - 1);
        
        if (isMax && (best == null || best.heuristic < child.heuristic)) {
          best = child;
          if (best.heuristic > parent.alphaBetaMin) {
            parent.alphaBetaMin = best.heuristic;
            System.out.println("Alpha-Beta: Max node bounds being set to (" + parent.alphaBetaMin + " / " + parent.alphaBetaMax + ")");
            if (parent.parent != null && parent.alphaBetaMin >= parent.parent.alphaBetaMax) {
              System.out.println("Alpha-Beta: Pruning remaining");
              break;
            }
          }
        } else if (!isMax && (best == null || best.heuristic > child.heuristic)) {
          best = child;
          if (best.heuristic < parent.alphaBetaMax) {
            parent.alphaBetaMax = best.heuristic;
            System.out.println("Alpha-Beta: Min node bounds being set to (" + parent.alphaBetaMin + " / " + parent.alphaBetaMax + ")");
            if (parent.parent != null && parent.alphaBetaMax <= parent.parent.alphaBetaMin) {
              System.out.println("Alpha-Beta: Pruning remaining");
              break;
            }
          }
        }
      }
      System.out.println("Returning from depth " + depth + " with heuristic: " + best.heuristic);
      System.out.println();
      
      return best;
    }
  }
  
  // Beam search to find a set of local maxima or minima
  public static TreeSet <Node> evaluateTurn(Bullies game, int beamWidth, boolean max, Player evaluator) {
    TreeSet <Node> openStates = new TreeSet <Node>();
    
    Player current = game.getActivePlayer();
    State root = State.push(game, current);
    openStates.add(new Node(root, State.evaluate(game, evaluator)));
    int basisHeuristic = max ? Integer.MIN_VALUE : Integer.MAX_VALUE; // Highest or lowest accepted heuristic
    
    int loopCounter = 0;
    boolean unchangedIteration = false; // openStates hasn't changed in the last loop
    while (!unchangedIteration) {
      unchangedIteration = true;
      
      for (Node open : new ArrayList <Node> (openStates)) {
        State.pop(game, open.state);
        ArrayList <Piece> active = game.getActivePieces();
        
        for (int i=0; i < active.size(); ++i) {
          for (Direction dir : Direction.values()) {
            game.move(active.get(i), dir);
            int heuristic = State.evaluate(game, evaluator);
            Node n = new Node(State.push(game, current), heuristic);
            
            if (!openStates.contains(n) && max ? heuristic > basisHeuristic : heuristic < basisHeuristic) {
              openStates.add(n);
              unchangedIteration = false;
              
              // if beam size has been reached then clamp set and redefine minimum
              if (openStates.size() > beamWidth) {
                if (max) {
                  openStates.pollFirst();
                  basisHeuristic = openStates.first().heuristic;
                } else {
                  openStates.pollLast();
                  basisHeuristic = openStates.last().heuristic;
                }
              }
            }
            
            State.pop(game, open.state);
            active = game.getActivePieces();
          }
        }
      }
      
      ++loopCounter;
      if (loopCounter > MAX_BEAM_DEPTH) break;
    }
    
    return openStates;
  }
  
  private static class Node implements Comparable <Node>{
    public State state;
    public Node parent;
    public int heuristic;
    public int depth = 0;
    public int alphaBetaMax = Integer.MAX_VALUE;
    public int alphaBetaMin = Integer.MIN_VALUE;
    
    protected Node(Bullies game) {
      this.state = State.push(game, game.getActivePlayer());
      this.heuristic = State.evaluate(game, game.getActivePlayer());
    }
    
    protected Node(State state, int heuristic) {
      this.state = state;
      this.heuristic = heuristic;
    }
    
    public int compareTo(Node open) {
      // HACK: Prevents the set's removal of duplicates by uniquely identifying all entries, but having the 
      // dominant characteristic still be the heuristics.
      // TODO: It would be preferable to prevent duplicates to improve performance
      int comparison = this.heuristic - open.heuristic;
//      /*
      comparison *= 10000;
      comparison += Math.random() * 10000;
//      */
      return comparison;
    }
    
    public boolean equals(Object o) {
      if (this == o) return true;
      else if (o instanceof Node) {
        Node op = (Node) o;
        return op.state.equals(this.state);
      } else return false;
    }
  }
}
