package game;

import java.awt.*;
import java.awt.event.*;
import java.awt.image.BufferedImage;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;

import javax.imageio.ImageIO;
import javax.swing.*;

import util.ComponentTools;
import util.Direction;
import util.squareGrid.Board;
import util.squareGrid.BoundsException;
import util.squareGrid.CollisionException;
import util.squareGrid.Placeable;

public class Bullies {
  private static final String IMG_BASE_PATH = "images/";
  private ArrayList <Player> participants = new ArrayList <Player> (); // Players with queens remaining
  private Player activePlayer;
  private Board board;
  private StarterDialog.Output dialogSettings = null;
  private GamePanel display;
  private boolean gameEnded = true;
  private boolean suppressGui = false; // Prevents gui events when AI's running
  private int minimaxDepth = 1;
  
  public static void main(String[] args) {
    final Bullies game = new Bullies();
    
    final JFrame window = new JFrame();
    window.setFocusable(false);
    window.setTitle("Bullies");
    window.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
    window.setResizable(false);
    window.add(game.getDisplay());
    
    // New game button
    JButton newgame = new JButton("New Game");
    newgame.setFocusable(false);
    newgame.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent event) {
        game.newGame(window);
      }
    });
    
    JPanel bottom = new JPanel(new FlowLayout(FlowLayout.LEFT));
    bottom.add(newgame);
    window.add(bottom, BorderLayout.SOUTH);
    
    window.pack();
    ComponentTools.center(window);
    window.setVisible(true);
  }
  
  public Bullies() {
    this.board = new Board(10, 10, false);
    this.display = new GamePanel(this);
  }
  
  // frame = null if being ran as an applet
  public void newGame(JFrame frame) {
    StarterDialog.Output output;
    if (this.dialogSettings != null) output = StarterDialog.showDialog(frame, this.dialogSettings);
    else output = StarterDialog.showDialog(frame);
    
    if (output != null) {
      this.gameEnded = false;
      this.dialogSettings = output;
      this.minimaxDepth = output.minimaxDepth;
      
      // Sets piece properties
      Piece.Type.DRONE.setMovement(output.droneMovement);
      Piece.Type.DRONE.setShoveDistance(output.droneShove);
      Piece.Type.QUEEN.setMovement(output.queenMovement);
      Piece.Type.QUEEN.setShoveDistance(output.queenShove);
      
      ArrayList <Piece> contents = new ArrayList <Piece> ();
      for (StarterDialog.PlayerOutput player : output.players) {
        if (player.getType() == StarterDialog.PlayerType.NONE) continue;
        
        Player.Type type = player.getType() == StarterDialog.PlayerType.HUMAN ? Player.Type.HUMAN : Player.Type.COMPUTER;
        Player newPlayer = new Player(player.getName(), player.getColor(), type);
        contents.addAll(player.getPieces(newPlayer, output.boardSize.width, output.boardSize.height));
      }
      
      this.board.clear();
      this.participants.clear();
      this.board = new Board(output.boardSize.width, output.boardSize.height, false);
      for (Piece piece : contents) {
        // Adds player to current list of participants in the game
        if (!this.participants.contains(piece.getOwner())) {
          this.participants.add(piece.getOwner());
        }
        
        this.board.add(piece);
      }
      
      this.activePlayer = this.participants.get(0);
      this.display.reset();
      this.display.repaint();
    }
  }
  
  public ArrayList <Player> getParticipants() {
    return new ArrayList <Player> (this.participants);
  }
  
  // Used by AI to set state
  protected void setState(ArrayList <Piece> pieces, Player active, ArrayList <Player> participants) {
    this.board.clear();
    this.participants.clear();
    for (Piece piece : pieces) {
      /*
      if (!this.participants.contains(piece.getOwner())) {
        this.participants.add(piece.getOwner());
      }
      */
      this.board.add(piece);
    }
    this.participants = new ArrayList <Player> (participants);
    
    this.activePlayer = active;
  }
  
  /*
  protected void setState(LinkedList <State.PieceTemplate> pieces, Player active) {
    this.participants.clear();
    Iterator <State.PieceTemplate> itr = pieces.iterator();
    State.PieceTemplate tempate = itr.next();
    
    for (Piece piece : getPieces()) {
      
      
      
      
      for (State.PieceTemplate template : pieces) {
        if (piece.getOwner().equals(template.getOwner()) && piece.getType().equals(template.getType())) {
          template.overwriteProperties(piece);
          break;
        }
      }
      
      
      
    }
  }
  */
  public GamePanel getDisplay() {
    return this.display;
  }
  
  // Casts all contents to pieces
  public ArrayList <Piece> getPieces() {
    ArrayList <Piece> pieces = new ArrayList <Piece> ();
    for (Placeable temp : this.board.getContents()) {
      pieces.add((Piece) temp);
    }
    return pieces;
  }
  
  // Gets all pieces belonging to a given player
  public ArrayList <Piece> getPieces(Player owner) {
    ArrayList <Piece> pieces = new ArrayList <Piece> ();
    for (Placeable temp : this.board.getContents()) {
      Piece piece = (Piece) temp;
      if (piece.getOwner().equals(owner)) pieces.add(piece);
    }
    return pieces;
  }
  
  // Provides a list of pieces currently able to move
  public ArrayList <Piece> getActivePieces() {
    ArrayList <Piece> active = new ArrayList <Piece> ();
    for (Placeable temp : this.board.getContents()) {
      Piece piece = (Piece) temp;
      if (piece.getOwner().equals(this.activePlayer) && piece.getMovement() > 0) active.add(piece);
    }
    return active;
  }
  
  public Player getActivePlayer() {
    return this.activePlayer;
  }
  
  public void move(Piece piece, Direction dir) {
    if (this.gameEnded) return;
    
    if (piece.getOwner().equals(this.activePlayer) && piece.getMovement() > 0) {
      try {
        piece.setDirection(dir);
        piece.move(dir);
        piece.setMovement(piece.getMovement() - 1);
      } catch (CollisionException exc) {
        if (piece.getType().getShoveDistance() < 1) return;
        
        // Piece performs a shove
        piece.setMovement(0);
        Point nextLoc = dir.getOffset(dir.getOffset(piece.getAnchor()));
        if (!this.board.isEmpty(nextLoc)) {
          // Group shove
          Point endingLoc = dir.getOffset(piece.getAnchor());
          while (!this.board.isEmpty(endingLoc)) {
            endingLoc = dir.getOffset(endingLoc);
          }
          endingLoc = dir.reverse().getOffset(endingLoc);
          
          // Moves each piece in line, starting with the end, one space
          while (!endingLoc.equals(piece.getAnchor())) {
            Piece occupant = (Piece) this.board.getLocation(endingLoc);
            if (occupant == null) break;
            
            try {
              occupant.move(dir);
            } catch (BoundsException exc2) {
              remove(occupant);
            } catch (CollisionException exc2) {
              // Shouldn't happen
              assert(false): "Collision detected in shove when there shouldn't be";
            }
            
            endingLoc = dir.reverse().getOffset(endingLoc);
          }
        } else {
          // Push shove distance or until it hits another piece
          Piece hit = (Piece) this.board.getLocation(dir.getOffset(piece.getAnchor()));
          try {
            for (int i = 0; i < piece.getType().getShoveDistance(); ++i) {
              hit.move(dir);
            }
          } catch (BoundsException exc2) {
            remove(hit);
          } catch (CollisionException exc2) {
            // Piece stops moving
          }
        }
      } catch (BoundsException exc) {
        remove(piece); // Suicide- piece is lost
      }
    }
  }
  
  public boolean remove(Placeable placeable) {
    boolean result = this.board.remove(placeable);
    Piece removed = (Piece) placeable;
    
    if (removed.getType() == Piece.Type.QUEEN) {
      // Checks if player has queens remaining
      ArrayList <Piece> alliedPieces = getPieces(removed.getOwner());
      boolean hasQueen = false;
      for (Piece piece : alliedPieces) {
        if (piece.getType() == Piece.Type.QUEEN) {
          hasQueen = true;
          break;
        }
      }
      
      if (!hasQueen) {
        // Game over for player - shifts active as a hack to keep ordering in the case of suicide
        int playerNum = (this.participants.indexOf(this.activePlayer) + this.participants.size() - 1) % this.participants.size();
        this.activePlayer = this.participants.get(playerNum);
        for (Piece piece : getPieces(this.activePlayer)) {
          piece.setMovement(0);
        }
        
        // Removes remaining pieces and player from participants
        this.participants.remove(removed.getOwner());
        for (Piece piece : alliedPieces) {
          this.board.remove(piece);
        }
        
        if (!this.suppressGui && this.participants.size() < 2) {
          this.display.repaint();
          this.gameEnded = true;
          
          if (this.participants.size() == 1) {
            Player winner = this.participants.get(0);
            JOptionPane.showMessageDialog(this.display, winner.getName() + " wins!");
          } else if (!this.suppressGui && this.participants.size() < 1) {
            // Tie game
            JOptionPane.showMessageDialog(this.display, "The game has ended in a draw.");
          }
        }
      }
    }
    
    return result;
  }
  
  // Initiates the next player's turn
  public void nextTurn() {
    if (this.gameEnded || this.participants.size() == 0) return;
    int playerNum = (this.participants.indexOf(this.activePlayer) + 1) % this.participants.size();
    this.activePlayer = this.participants.get(playerNum);
    
    // Resets movement of pieces
    for (Piece piece : getPieces(this.activePlayer)) {
      piece.resetMovement();
    }
    
    // Uses AI if computer controlled
    // Gui suppression is checked to see if this is an AI emulation
    if (this.activePlayer.getType() == Player.Type.COMPUTER && !this.suppressGui) {
      this.suppressGui = true;
      State newState = AITools.getNextMove(this, this.minimaxDepth);
      State.pop(this, newState);
      this.suppressGui = false;
    }
  }
  
  public Board getBoard() {
    return this.board;
  }
  
  // Gets an image from the image folder
  protected static BufferedImage getImageResource(String resource) throws IOException {
    String path = IMG_BASE_PATH + resource;
    InputStream stream = Piece.class.getResourceAsStream(path);
    if (stream == null) throw new IOException("Image resource not found: " + path);
    BufferedImage img = ImageIO.read(stream);
    stream.close();
    return img;
  }
}
