package game;

import java.awt.*;
import java.awt.event.*;
import java.awt.image.BufferedImage;
import java.io.IOException;
import java.util.ArrayList;

import javax.swing.*;

import util.Direction;
import util.media.Stamp;

/**
 * User interface reflecting the current state of a game, and provides controls for human players.
 * @author Damian Johnson (atagar1@gmail.com)
 * @version October 21, 2007
 */
public class GamePanel extends JPanel {
  public static final long serialVersionUID = 0;
  public JApplet appletReference = null; // Hack to get #*&@ keybindings working
  private static final Dimension WINDOW_SIZE = new Dimension(600, 600);
  private static final Color BLANK_COLOR = Color.BLACK;
  private int squareSize = 70;
  private static final BufferedImage BACKGROUND;
  private static final BufferedImage BLANK;
  private static final boolean DRAW_GRID = true;
  private static final Color GRID_COLOR = Color.BLACK;
  private Bullies game;
  private Piece selected = null; // Null if none is selected
  
  static {
    String backgroundPath = "background.png";
    String blankPath = "blank.png";
    
    BufferedImage backgroundImg;
    BufferedImage blankImg;
    try {
      backgroundImg = Bullies.getImageResource(backgroundPath);
      blankImg = Bullies.getImageResource(blankPath);
    } catch (IOException exc) {
      backgroundImg = null;
      blankImg = null;
      System.out.println(exc.getMessage());
    }
    
    BACKGROUND = backgroundImg;
    BLANK = blankImg;
  }
  
  public GamePanel(Bullies bullies) {
    this.game = bullies;
    this.setPreferredSize(WINDOW_SIZE);
    reset();
    
    this.addMouseListener(new MouseAdapter() {
      public void mouseClicked(MouseEvent event) {
        if (appletReference != null) {
          appletReference.requestFocus();
          appletReference.requestFocusInWindow();
        }
        
        int x = event.getPoint().x / squareSize;
        int y = event.getPoint().y / squareSize;
        Point loc = new Point(x, y);
        
        if (game.getBoard().isEmpty(loc)) {
          selected = null;
        } else {
          selected = (Piece) game.getBoard().getLocation(loc);
        }
        
        repaint();
      }
    });
    
    // These controls are ignored when being ran as an applet
    InputMap input = getInputMap();
    input.put(KeyStroke.getKeyStroke(KeyEvent.VK_UP, 0), "Move Up");
    input.put(KeyStroke.getKeyStroke(KeyEvent.VK_RIGHT, 0), "Move Right");
    input.put(KeyStroke.getKeyStroke(KeyEvent.VK_DOWN, 0), "Move Down");
    input.put(KeyStroke.getKeyStroke(KeyEvent.VK_LEFT, 0), "Move Left");
    input.put(KeyStroke.getKeyStroke(KeyEvent.VK_ENTER, 0), "Next Turn");
    
    ActionMap actions = getActionMap();
    actions.put("Move Up", new MovementEvent(Direction.UP));
    actions.put("Move Right", new MovementEvent(Direction.RIGHT));
    actions.put("Move Down", new MovementEvent(Direction.DOWN));
    actions.put("Move Left", new MovementEvent(Direction.LEFT));
    actions.put("Next Turn", new AbstractAction() {
      public static final long serialVersionUID = 0;
      public void actionPerformed(ActionEvent event) {
        game.nextTurn();
        repaint();
      }
    });
  }
  
  // Resets display properties for new games
  public void reset() {
    int w = WINDOW_SIZE.width / this.game.getBoard().getWidth();
    int h = WINDOW_SIZE.height / this.game.getBoard().getHeight();
    this.squareSize = Math.min(w, h);
  }
  
  public void paint(Graphics g) {
    Graphics2D g2 = (Graphics2D) g;
    Dimension dim = getPreferredSize();
    
    // Fills the background
    if (BACKGROUND != null) {
      for (int width = 0; width < dim.width; width += BACKGROUND.getWidth()) {
        for (int height = 0; height < dim.height; height += BACKGROUND.getHeight()) {
          g2.drawImage(BACKGROUND, width, height, null);
        }
      }
    } else {
      g2.setColor(Color.WHITE);
      g2.fillRect(0, 0, dim.width, dim.height);
    }
    
    // Fills unused space
    int contentWidth = this.game.getBoard().getWidth() * this.squareSize;
    int contentHeight = this.game.getBoard().getHeight() * this.squareSize;
    Dimension contentSize = new Dimension(contentWidth, contentHeight);
    g2.setColor(BLANK_COLOR);
    if (contentWidth < WINDOW_SIZE.width) {
      int blankWidth = WINDOW_SIZE.width - contentWidth;
      if (BLANK != null) {
        for (int w = 0; w < blankWidth; w += BLANK.getWidth()) {
          for (int h = 0; h < WINDOW_SIZE.height; h += BLANK.getHeight()) {
            g2.drawImage(BLANK, contentWidth + w, h, null);
          }
        }
      } else g2.fillRect(contentWidth, 0, blankWidth, WINDOW_SIZE.height);
    }
    
    if (contentHeight < WINDOW_SIZE.height) {
      int blankHeight = WINDOW_SIZE.height - contentHeight;
      if (BLANK != null) {
        for (int w = 0; w < WINDOW_SIZE.width; w += BLANK.getWidth()) {
          for (int h = 0; h < contentHeight; h += BLANK.getHeight()) {
            g2.drawImage(BLANK, w, contentHeight + h, null);
          }
        }
      } else g2.fillRect(0, contentHeight, WINDOW_SIZE.width, blankHeight);
    }
    
    // Draws grid
    if (DRAW_GRID) {
      g2.setColor(GRID_COLOR);
      for (int row = 0; row < this.game.getBoard().getHeight() + 1; ++row) {
        int y = row * squareSize;
        g2.drawLine(0, y, contentSize.width, y);
      }
      
      for (int col = 0; col < this.game.getBoard().getWidth() + 1; ++ col) {
        int x = col * squareSize;
        g2.drawLine(x, 0, x, contentSize.height);
      }
    }
    
    // Draws contents
    ArrayList <Piece> active = this.game.getActivePieces();
    for (Piece piece : this.game.getPieces()) {
      int x = piece.getAnchor().x * squareSize;
      int y = piece.getAnchor().y * squareSize;
      g2.drawImage(piece.getImage(squareSize), x, y, null);
      
      // Notes remaining movement
      if (active.contains(piece)) {
        g2.setColor(Color.BLACK);
        int halfSquare = squareSize / 2;
        g2.fillRect(x + halfSquare, y, halfSquare, halfSquare);
        
        g2.setColor(piece.getOwner().getColor());
        g2.setFont(new Font("Monospaced", Font.BOLD, halfSquare));
        g2.drawString("" + piece.getMovement(), x + squareSize * 4 / 7, y + squareSize * 3 / 7);
      }
      
      // Brackets selected
      if (this.selected != null && this.selected.equals(piece)) {
        g2.setStroke(new BasicStroke(3, BasicStroke.CAP_ROUND, BasicStroke.JOIN_ROUND));
        
        Color playerColor = this.selected.getOwner().getColor();
        Stamp brackets = new Stamp(Stamp.Shape.CORNER, playerColor, squareSize);
        brackets.setScale(0.75);
        brackets.borderCorners(g2, new Point(x, y), false);
      }
    }
  }
  
  // Part of the hack to emulate input/action map for applet
  public void callMove(Direction dir) {
    if (selected == null || !game.getActivePieces().contains(selected)) return;
    else {
      game.move(selected, dir);
      repaint();
    }
  }
  
  private class MovementEvent extends AbstractAction {
    public static final long serialVersionUID = 0;
    private Direction dir;
    
    MovementEvent(Direction dir) {
      this.dir = dir;
    }
    
    public void actionPerformed(ActionEvent event) {
      if (selected == null || !game.getActivePieces().contains(selected)) return;
      else {
        game.move(selected, this.dir);
        repaint();
      }
    }
  }
}
