package game;

import java.awt.Color;
import java.awt.Point;
import java.awt.image.BufferedImage;
import java.io.IOException;
import java.util.HashMap;

import util.Direction;
import util.StringTools;
import util.media.GraphicTools;
import util.squareGrid.Placeable;

/**
 * Game piece for the Bullies game.
 * @author Damian Johnson (atagar1@gmail.com)
 * @version October 21, 2007
 */
public class Piece extends Placeable {
  private Type type;
  private Player owner;
  private int movement;
  private static HashMap <Color, HashMap <Direction, BufferedImage>> droneSets = 
    new HashMap <Color, HashMap <Direction, BufferedImage>> ();
  private static HashMap <Color, HashMap <Direction, BufferedImage>> queenSets = 
    new HashMap <Color, HashMap <Direction, BufferedImage>> ();
  
  public Piece(Point location, Type type, Player owner) {
    super(location);
    this.type = type;
    this.owner = owner;
    this.movement = this.type.getMovement();
    
    HashMap <Color, HashMap <Direction, BufferedImage>> imageSet;
    if (type == Type.DRONE) imageSet = droneSets;
    else imageSet = queenSets;
    
    if (imageSet.containsKey(this.owner.getColor())) {
      setImageSet(imageSet.get(this.owner.getColor()));
    } else {
      setImage(this.type.getImage(this.owner.getColor()));
      imageSet.put(this.owner.getColor(), getImageSet());
    }
  }
  
  public Piece(Piece piece) {
    this(new Point(piece.getAnchor()), piece.type, piece.owner);
    this.movement = piece.movement;
    setDirection(piece.getDirection());
  }
  
  public Player getOwner() {
    return this.owner;
  }
  
  public Type getType() {
    return this.type;
  }
  
  public int getMovement() {
    return this.movement;
  }
  
  public void resetMovement() {
    this.movement = this.type.getMovement();
  }
  
  public void setMovement(int movement) {
    this.movement = movement;
  }
  
  public boolean equals(Object o) {
    if (this == o) return true;
    else if (o instanceof Piece) {
      Piece p = (Piece) o;
      boolean equals = p.getType().equals(getType());
      equals &= p.getOwner().equals(getOwner());
      equals &= p.getAnchor().equals(getAnchor());
      return equals;
    } else return false;
  }
  
  public String toString() {
    StringBuilder builder = new StringBuilder();
    if (this.type == Type.DRONE) builder.append("D (");
    else builder.append("Q (");
    
    builder.append(getAnchor().x);
    builder.append("/");
    builder.append(getAnchor().y);
    
    builder.append(" O: ");
    builder.append(this.owner.getName());
    
    builder.append(" M: ");
    builder.append(this.movement);
    builder.append(")");
    return builder.toString();
  }
  
  public static enum Type {
    DRONE("drone.png"), QUEEN("queen.png");
    
    private BufferedImage pieceImage; // Cached image of piece type
    private HashMap <Color, BufferedImage> playerImages = new HashMap <Color, BufferedImage> ();
    private int movement = 5;
    private int shove = 3;
    private Type(String imagePath) {
      try {
        this.pieceImage = Bullies.getImageResource(imagePath);
      } catch (IOException exc) {
        System.out.println(exc.getMessage());
      }
    }
    
    public int getMovement() {
      return this.movement;
    }
    
    public void setMovement(int movement) {
      this.movement = movement;
    }
    
    public int getShoveDistance() {
      return this.shove;
    }
    
    public void setShoveDistance(int shove) {
      this.shove = shove;
    }
    
    public BufferedImage getImage() {
      return GraphicTools.copy(this.pieceImage);
    }
    
    public BufferedImage getImage(Color color) {
      // WARNING- this passes the image rather than a duplicate
      if (this.playerImages.containsKey(color)) {
        return this.playerImages.get(color);
      }
      else {
        BufferedImage pieceImage = getImage();
        GraphicTools.replaceColor(pieceImage, Color.WHITE, color);
        this.playerImages.put(color, pieceImage);
        return pieceImage;
      }
    }
    
    public String toString() {
      return StringTools.getReadableConstant(this.name());
    }
  }
}
