package game;

import java.awt.*;
import java.awt.event.*;
import java.awt.image.BufferedImage;
import java.io.IOException;
import java.text.NumberFormat;
import java.util.ArrayList;

import javax.swing.*;
import javax.swing.text.NumberFormatter;

import util.ComponentTools;
import util.Corner;
import util.media.GraphicTools;

/**
 * Dialog prompting input for a new session of the Bullies game.
 * @author Damian Johnson (atagar1@gmail.com)
 * @version December 5, 2007
 */
class StarterDialog extends JDialog {
  public static final long serialVersionUID = 0;
  private static final String TITLE = "New Game";
  
  private static final int PLAYER_COUNT = 4;
  private static final Color PLAYER1_COLOR = Color.RED;
  private static final Color PLAYER2_COLOR = Color.BLUE;
  private static final Color PLAYER3_COLOR = Color.ORANGE;
  private static final Color PLAYER4_COLOR = Color.GRAY;
  private static final int DEFAULT_DRONES = 5;
  private static final int DEFAULT_QUEENS = 1;
  private static final int POSITION_OFFSET = 1; // Minimum initial piece distance from border
  
  private JFormattedTextField widthField;
  private JFormattedTextField heightField;
  private JFormattedTextField droneMovement;
  private JFormattedTextField droneShove;
  private JFormattedTextField queenMovement;
  private JFormattedTextField queenShove;
  private JFormattedTextField minimaxDepth;
  private ArrayList <PlayerOutput> players = new ArrayList <PlayerOutput> ();
  private boolean okPressed = false;
  
  {
    NumberFormatter doubleDigitFormat =
      new NumberFormatter(NumberFormat.getIntegerInstance());
    doubleDigitFormat.setMinimum(1);
    doubleDigitFormat.setMaximum(99);
    
    this.widthField = new JFormattedTextField(doubleDigitFormat);
    this.heightField = new JFormattedTextField(doubleDigitFormat);
    this.droneMovement = new JFormattedTextField(doubleDigitFormat);
    this.droneShove = new JFormattedTextField(doubleDigitFormat);
    this.queenMovement = new JFormattedTextField(doubleDigitFormat);
    this.queenShove = new JFormattedTextField(doubleDigitFormat);
    this.minimaxDepth = new JFormattedTextField(doubleDigitFormat);
    
    this.widthField.setColumns(2);
    this.heightField.setColumns(2);
    this.droneMovement.setColumns(2);
    this.droneShove.setColumns(2);
    this.queenMovement.setColumns(2);
    this.queenShove.setColumns(2);
    this.minimaxDepth.setColumns(2);
    
    this.droneMovement.setHorizontalAlignment(JTextField.RIGHT);
    this.droneShove.setHorizontalAlignment(JTextField.RIGHT);
    this.queenMovement.setHorizontalAlignment(JTextField.RIGHT);
    this.queenShove.setHorizontalAlignment(JTextField.RIGHT);
  }
  
  protected static Output showDialog(Frame parent) {
    return showDialog(parent, new Output());
  }
  
  protected static Output showDialog(Frame parent, Output defaults) {
    StarterDialog dialog = new StarterDialog(parent, defaults);
    if (parent != null) ComponentTools.center(dialog, parent);
    else ComponentTools.center(dialog);
    dialog.setVisible(true);
    
    if (dialog.okPressed) {
      Output output = new Output();
      Dimension boardSize = new Dimension();
      boardSize.width = ((Number) dialog.widthField.getValue()).intValue();
      boardSize.height = ((Number) dialog.heightField.getValue()).intValue();
      output.boardSize = boardSize;
      
      output.droneMovement = ((Number) dialog.droneMovement.getValue()).intValue();
      output.droneShove = ((Number) dialog.droneShove.getValue()).intValue();
      output.queenMovement = ((Number) dialog.queenMovement.getValue()).intValue();
      output.queenShove = ((Number) dialog.queenShove.getValue()).intValue();
      output.players = dialog.players;
      output.minimaxDepth = ((Number) dialog.minimaxDepth.getValue()).intValue();
      return output;
    } else {
      return null;
    }
  }
  
  private StarterDialog(Frame parent, Output defaults) {
    super(parent, true);
    setTitle(TITLE);
    setDefaultCloseOperation(JDialog.DISPOSE_ON_CLOSE);
    setResizable(false);
    setLayout(new BorderLayout());
    
    // Players section
    JPanel playersSection = new JPanel(new GridLayout(2, 2, 0, 10));
    for (PlayerOutput temp : defaults.players) {
      PlayerOutput player = new PlayerOutput(temp);
      this.players.add(player);
      playersSection.add(makePlayerSection(player));
    }
    add(playersSection, BorderLayout.WEST);
    
    // Board section
    JPanel right = new JPanel(new BorderLayout());
    JPanel top = new JPanel(new BorderLayout());
    JPanel boardSection = new JPanel(new FlowLayout(FlowLayout.LEFT));
    boardSection.setBorder(BorderFactory.createTitledBorder("Board"));
    
    boardSection.add(new JLabel("Width: "));
    this.widthField.setValue(defaults.boardSize.width);
    boardSection.add(this.widthField);
    
    boardSection.add(new JLabel("     ")); // Spacer
    
    boardSection.add(new JLabel("Height: "));
    this.heightField.setValue(defaults.boardSize.height);
    boardSection.add(this.heightField);
    top.add(boardSection, BorderLayout.NORTH);
    
    // AI section
    JPanel aiSection = new JPanel(new FlowLayout(FlowLayout.LEFT));
    aiSection.setBorder(BorderFactory.createTitledBorder("AI"));
    aiSection.add(new JLabel("Minimax Depth: "));
    this.minimaxDepth.setValue(defaults.minimaxDepth);
    aiSection.add(this.minimaxDepth);
    top.add(aiSection, BorderLayout.SOUTH);
    right.add(top, BorderLayout.NORTH);
    
    // Units section
    Box unitsSection = Box.createVerticalBox();
    unitsSection.setBorder(BorderFactory.createTitledBorder("Units"));
    
    this.droneMovement.setValue(defaults.droneMovement);
    this.droneShove.setValue(defaults.droneShove);
    unitsSection.add(makeUnitSection(Piece.Type.DRONE, this.droneMovement, this.droneShove));
    
    this.queenMovement.setValue(defaults.queenMovement);
    this.queenShove.setValue(defaults.queenShove);
    unitsSection.add(makeUnitSection(Piece.Type.QUEEN, this.queenMovement, this.queenShove));
    right.add(unitsSection, BorderLayout.CENTER);
    
    JPanel controls = new JPanel(new FlowLayout(FlowLayout.LEFT));
    JButton ok = new JButton("Ok");
    ok.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent event) {
        ArrayList <String> errors = getErrors();
        if (errors.size() > 0) {
          boolean plural = errors.size() > 1;
          StringBuilder message = new StringBuilder();
          message.append("Problem");
          if (plural) message.append("s");
          
          message.append(" encountered in starting the game:\n");
          for (int i=0; i < errors.size(); ++i) {
            if (plural) message.append((i+1) + ". ");
            message.append(errors.get(i));
          }
          
          JOptionPane.showMessageDialog(StarterDialog.this, message.toString(), "Problems", JOptionPane.ERROR_MESSAGE);
        } else {
          okPressed = true;
          StarterDialog.this.dispose();
        }
      }
    });
    controls.add(ok);
    
    JButton cancel = new JButton("Cancel");
    cancel.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent event) {
        okPressed = false;
        StarterDialog.this.dispose();
      }
    });
    controls.add(cancel);
    right.add(controls, BorderLayout.SOUTH);
    
    JPanel wrapper = new JPanel();
    wrapper.add(right);
    add(wrapper, BorderLayout.EAST);
    
    pack();
  }
  
  private JPanel makeUnitSection(Piece.Type type, JFormattedTextField movement, JFormattedTextField shove) {
    JPanel section = new JPanel(new BorderLayout());
    BufferedImage img = type.getImage();
    GraphicTools.replaceColor(img, Color.WHITE, Color.RED);
    section.add(new JLabel(" ", new ImageIcon(img), SwingConstants.LEFT), BorderLayout.WEST);
    
    JPanel content = new JPanel(new BorderLayout());
    JLabel label = new JLabel(type.toString());
    label.setFont(new Font("SansSerif", Font.BOLD, 24));
    content.add(label, BorderLayout.NORTH);
    
    JPanel optionLabels = new JPanel(new GridLayout(0, 1));
    JPanel options = new JPanel(new GridLayout(0, 1));
    JPanel wrapper = new JPanel(new BorderLayout());
    optionLabels.add(new JLabel("Movement: "));
    options.add(movement);
    
    optionLabels.add(new JLabel("Shove: "));
    options.add(shove);
    
    wrapper.add(optionLabels, BorderLayout.WEST);
    wrapper.add(options, BorderLayout.EAST);
    
    JPanel outerWrapper = new JPanel();
    outerWrapper.add(wrapper);
    content.add(outerWrapper);
    
    section.add(content, BorderLayout.EAST);
    return section;
  }
  
  private JPanel makePlayerSection(final PlayerOutput player) {
    JPanel section = new JPanel(new BorderLayout());
    
    JLabel label = new JLabel(" " + player.getName());
    label.setForeground(player.getColor());
    label.setFont(new Font("SansSerif", Font.BOLD, 20));
    
    section.add(label, BorderLayout.NORTH);
    
    JButton playerType = new JButton(new ImageIcon(player.getType().getImage()));
    playerType.setFocusable(false); // Prevents focus box
    playerType.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent event) {
        JButton source = (JButton) event.getSource();
        player.setType(player.getType().next());
        source.setIcon(new ImageIcon(player.getType().getImage()));
      }
    });
    section.add(playerType, BorderLayout.CENTER);
    
    JPanel bottom = new JPanel(new BorderLayout());
    
    JPanel optionLabels = new JPanel(new GridLayout(0, 1));
    optionLabels.add(new JLabel(" Drones: "));
    optionLabels.add(new JLabel(" Queens: "));
    bottom.add(optionLabels, BorderLayout.WEST);
    
    JPanel wrapper = new JPanel(new BorderLayout());
    JPanel options = new JPanel(new GridLayout(0, 1));
    wrapper.add(player.getDronesField(), BorderLayout.EAST);
    options.add(wrapper);
    
    wrapper = new JPanel(new BorderLayout());
    wrapper.add(player.getQueensField());
    options.add(wrapper);
    bottom.add(options, BorderLayout.EAST);
    
    section.add(bottom, BorderLayout.SOUTH);
    return section;
  }
  
  private ArrayList <String> getErrors() {
    ArrayList <String> errors = new ArrayList <String> ();
    
    // Checks that each player has a queen
    int playerCount = 0;
    for (PlayerOutput player : this.players) {
      if (player.getType() != PlayerType.NONE) {
        ++playerCount;
        
        // This condition is prevented by bounds on the field
        if (player.getQueens() < 1) {
          String message = player.getName() + " needs at least one queen to play.";
          errors.add(message);
        }
      }
    }
    
    // Checks that there's enough participants to play
    if (playerCount < 2) {
      errors.add("There must be at least two players.");
    }
    
    // Checks that there's enough room for initial pieces
    // TODO: This is a hack- resulting contents should be checked instead
    int occupiedTopWidth = POSITION_OFFSET * 2;
    int occupiedBottomWidth = POSITION_OFFSET * 2;
    
    int occupiedLeftHeight = POSITION_OFFSET * 2;
    int occupiedRightHeight = POSITION_OFFSET * 2;
    
    for (PlayerOutput player : this.players) {
      if (player.getType() == PlayerType.NONE) continue;
      else {
        int occupiedHeight = (int) Math.ceil(Math.sqrt(player.getDrones() + player.getQueens()));
        int occupiedWidth = (int) Math.round(Math.sqrt(player.getDrones() + player.getQueens()));
        
        if (player.initialCorner.isLeft()) occupiedLeftHeight += occupiedHeight;
        else occupiedRightHeight += occupiedHeight;
        
        if (player.initialCorner.isUpper()) occupiedTopWidth += occupiedWidth;
        else occupiedBottomWidth += occupiedWidth;
      }
    }
    
    if (Math.max(occupiedTopWidth, occupiedBottomWidth) > ((Number) widthField.getValue()).intValue()) {
      String message = "There's not enough room for pieces-\n     increase width or reduce initial pieces.";
      errors.add(message);
    }
    
    if (Math.max(occupiedLeftHeight, occupiedRightHeight) > ((Number) heightField.getValue()).intValue()) {
      String message = "There's not enough room for pieces-\n     increase height or reduce initial pieces.";
      errors.add(message);
    }
    
    return errors;
  }
  
  /**
   * Input given to this dialog.
   */
  static class Output {
    // Initial values are the defaults of the dialog
    public Dimension boardSize = new Dimension(10, 10);
    public int droneMovement = 5;
    public int droneShove = 3;
    public int queenMovement = 5;
    public int queenShove = 0;
    public int minimaxDepth = 1;
    public ArrayList <PlayerOutput> players;
    
    private Output() {
      this.players = new ArrayList <PlayerOutput> ();
      
      for (int i=0; i < PLAYER_COUNT; ++i) {
        Color playerColor;
        if (i == 0) playerColor = PLAYER1_COLOR;
        else if (i == 1) playerColor = PLAYER2_COLOR;
        else if (i == 2) playerColor = PLAYER3_COLOR;
        else if (i == 3) playerColor = PLAYER4_COLOR;
        else {
          assert(false): "No color defined for players beyond " + PLAYER_COUNT;
          playerColor = Color.BLACK;
        }
        
        PlayerOutput player = new PlayerOutput(i+1, playerColor);
        if (i > 1) player.setType(PlayerType.NONE);
        this.players.add(player);
      }
    }
  }
  
  /**
   * Input about a particular player given to this dialog.
   */
  static class PlayerOutput {
    private PlayerType type = PlayerType.HUMAN;
    private String name;
    private Color color;
    private JFormattedTextField dronesField;
    private JFormattedTextField queensField;
    private Corner initialCorner;
    
    private PlayerOutput(int playerNum, Color color) {
      this.name = "Player " + playerNum;
      this.color = color;
      
      NumberFormatter doubleDigitFormat =
        new NumberFormatter(NumberFormat.getIntegerInstance());
      doubleDigitFormat.setMinimum(1);
      doubleDigitFormat.setMaximum(99);
      
      this.dronesField = new JFormattedTextField(doubleDigitFormat);
      this.dronesField.setValue(DEFAULT_DRONES);
      this.dronesField.setColumns(2);
      this.dronesField.setHorizontalAlignment(JTextField.RIGHT);
      
      this.queensField = new JFormattedTextField(doubleDigitFormat);
      this.queensField.setValue(DEFAULT_QUEENS);
      this.queensField.setColumns(2);
      this.queensField.setHorizontalAlignment(JTextField.RIGHT);
      
      if (playerNum == 1) this.initialCorner = Corner.UPPER_LEFT;
      else if (playerNum == 2) this.initialCorner = Corner.UPPER_RIGHT;
      else if (playerNum == 3) this.initialCorner = Corner.LOWER_LEFT;
      else this.initialCorner = Corner.LOWER_RIGHT;
    }
    
    protected PlayerOutput (PlayerOutput initial) {
      this.type = initial.type;
      this.name = initial.name;
      this.color = initial.color;
      this.dronesField = initial.dronesField;
      this.queensField = initial.queensField;
      this.initialCorner = initial.initialCorner;
    }
    
    public PlayerType getType() {
      return this.type;
    }
    
    public void setType(PlayerType type) {
      this.type = type;
      
      boolean fieldsEnabled = getType() != PlayerType.NONE;
      getDronesField().setEnabled(fieldsEnabled);
      getQueensField().setEnabled(fieldsEnabled);
    }
    
    public Color getColor() {
      return this.color;
    }
    
    public String getName() {
      return this.name;
    }
    
    public int getDrones() {
      return ((Number) this.dronesField.getValue()).intValue();
    }
    
    JFormattedTextField getDronesField() {
      return this.dronesField;
    }
    
    public int getQueens() {
      return ((Number) this.queensField.getValue()).intValue();
    }
    
    JFormattedTextField getQueensField() {
      return this.queensField;
    }
    
    public ArrayList <Piece> getPieces(Player player, int width, int height) {
      ArrayList <Piece> pieces = new ArrayList <Piece> ();
      
      int w = (int) Math.round(Math.sqrt(getDrones() + getQueens()));
      int h = (int) Math.ceil(Math.sqrt(getDrones() + getQueens()));
      
      int dronesLeft = getDrones();
      int queensLeft = getQueens();
      for (int c=0; c < w; ++c) {
        for (int r=0; r < h; ++r) {
          int x = POSITION_OFFSET + c;
          if (!this.initialCorner.isLeft()) x = width - x - 1;
          
          int y = POSITION_OFFSET + r;
          if (!this.initialCorner.isUpper()) y = height - y - 1;
          
          Piece.Type type;
          Point loc = new Point(x, y);
          if (queensLeft > 0) {
            type = Piece.Type.QUEEN;
            --queensLeft;
          } else if (dronesLeft > 0) {
            type = Piece.Type.DRONE;
            --dronesLeft;
          } else break;
          
          pieces.add(new Piece(loc, type, player));
        }
      }
      
      return pieces;
    }
  }
  
  public enum PlayerType {
    NONE("player_none.png"),
    HUMAN("player_person.png"),
    COMPUTER("player_computer.png");
    
    private BufferedImage playerImage; // Cached image of player type
    private PlayerType(String imagePath) {
      try {
        this.playerImage = Bullies.getImageResource(imagePath);
      } catch (IOException exc) {
        System.out.println(exc.getMessage());
      }
    }
    
    public BufferedImage getImage() {
      return this.playerImage;
    }
    
    private PlayerType next() {
      if (this == NONE) return HUMAN;
      if (this == HUMAN) return COMPUTER;
      if (this == COMPUTER) return NONE;
      else throw new IllegalArgumentException("PlayerType method 'next()' doesn't recognize " + this);
    }
  }
}
