package util.media;
/** GraphicTools class
 * @author Damian Johnson
 * @version August 8, 2006
 */

import java.awt.*;
import java.awt.image.*;
import javax.swing.*;
import java.util.*;
import util.Direction;

/** Bundle of graphical utilities for altering the orientation,
 * coloring, and other aspects of images.
 */
public class GraphicTools {
    /**
     * Creates a buffered instance of an image of an 8-bit ARGB type.
     * @param image     picture to be processed
     * @return buffered instance of the image
     */
	public static BufferedImage buffer(Image image) {
		if (image instanceof BufferedImage) {
			return (BufferedImage) image;
		} else {
			int width = image.getWidth(null);
			int height = image.getHeight(null);
			BufferedImage buffered = new BufferedImage(width, height, BufferedImage.TYPE_INT_ARGB);
	        Graphics g = buffered.getGraphics();
	        g.drawImage(image, 0, 0, null);
	        return buffered;
		}
	}
	
    /**
     * Adds a component to a frame and displays it.
     * @param comp     component contents of frame
     * @return displayed frame
     */
	public static JFrame display(Component comp) {
		JFrame frame = new JFrame("Display");
		frame.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
		frame.add(comp);
    	frame.pack();
    	frame.setVisible(true);
		return frame;
	}
	
    /**
     * Adds an image to a frame and displays it.
     * @param image     image contents of frame
     * @return displayed frame
     */
	public static JFrame display(Image image) {
		JLabel label = new JLabel();
		label.setIcon(new ImageIcon(image));
		label.setBackground(Color.BLACK);
		label.setOpaque(true);
		return display(label);
	}
	
    /**
     * Compares the rgb values (including alpha) of two images to 
     * find if they're equal.
     * @param first     image to be compared
     * @param second     image to be compared
     * @return true if images are the same, false otherwise
     */
	public static boolean equals(BufferedImage first, BufferedImage second) {
		int width = first.getWidth();
		if (width != second.getWidth()) return false;
		int height = first.getHeight();
		if (height != second.getHeight()) return false;
		for (int row=0; row < width; ++row) {
			for (int col=0; col < height; ++col) {
				int rgb1 = first.getRGB(row, col);
				int rgb2 = second.getRGB(row, col);
				if (rgb1 != rgb2) return false;
			}
		}
		return true;
	}
	
    /**
     * Creates a deep copy of an image.
     * @param image     picture to be processed
     * @return copy of the image
     */
	public static BufferedImage copy(BufferedImage image) {
		int width = image.getWidth();
		int height = image.getHeight();
		BufferedImage copy = new BufferedImage(width, height, BufferedImage.TYPE_INT_ARGB);
		for (int row=0; row < width; ++row) {
			for (int col=0; col < height; ++col) {
				int rgb = image.getRGB(row, col);
				copy.setRGB(row, col, rgb);
			}
		}
        return copy;
	}
	
    /**
     * Reduces an image to the area between two points, 
     * inclusively.
     * @param image     picture to be processed
     * @param p1     bounds of area to crop
     * @param p2     bounds of area to crop
     * @return cropped instance of image
     */
	public static BufferedImage crop(BufferedImage image, Point p1, Point p2) {
		// Keeps the point values in the range of 0 - image dimension
		Dimension dim = new Dimension(image.getWidth(), image.getHeight());
		p1 = ensureRange(p1, new Point(0, 0), dim);
		p2 = ensureRange(p2, new Point(0, 0), dim);
		
		int width = Math.abs(p1.x - p2.x) + 1;
		int height = Math.abs(p1.y - p2.y) + 1;
		BufferedImage cropped = new BufferedImage(width, height, BufferedImage.TYPE_INT_ARGB);
		
		for (int row=0; row < image.getWidth(); ++row) {
			for (int col=0; col < image.getHeight(); ++col) {
				if (isBetween(p1, p2, new Point(row, col))) {
					int rgb = image.getRGB(row, col);
					int x = row - Math.min(p1.x, p2.x);
					int y = col - Math.min(p1.y, p2.y);
					cropped.setRGB(x, y, rgb);
				}
			}
		}
        return cropped;
	}
	
	// Returns a point with its coordinates within specific dimensions
	private static Point ensureRange(Point location, Point anchor, Dimension dim) {
		Point loc = new Point(location);
		loc.x = Math.max(loc.x, anchor.x);
		loc.x = Math.min(loc.x, anchor.x + dim.width);
		loc.y = Math.max(loc.y, anchor.y);
		loc.y = Math.min(loc.y, anchor.y + dim.height);
		return loc;
	}
	
	// Determines if the comparison point is within the rectangle 
	// made by the other points, inclusively
	private static boolean isBetween(Point p1, Point p2, Point comparison) {
		boolean btwnHor = comparison.x >= p1.x && comparison.x <= p2.x;
		btwnHor |= comparison.x <= p1.x && comparison.x >= p2.x;
		
		boolean btwnVer = comparison.y >= p1.y && comparison.y <= p2.y;
		btwnVer |= comparison.y <= p1.y && comparison.y >= p2.y;
		return btwnHor && btwnVer;
	}
	
	public static boolean inBounds(Point location, BufferedImage image) {
		boolean inBounds = true;
		inBounds &= location.x >= 0;
		inBounds &= location.x < image.getWidth();
		inBounds &= location.y >= 0;
		inBounds &= location.y < image.getHeight();
		return inBounds;
	}
	
	/**
	 * Centers a window within the screen.
     * @param frame     window to be centered
	 */
	public static void setCentered(Window frame) {
		Toolkit tk = Toolkit.getDefaultToolkit();
		Dimension screenSize = tk.getScreenSize();
		
		Dimension frameSize = frame.getSize();
		int x = (int) (screenSize.getWidth() / 2 - frameSize.getWidth() / 2);
		int y = (int) (screenSize.getHeight() / 2 - frameSize.getHeight() / 2);
		frame.setLocation(x, y);
	}
	
    /**
     * Factors a numeric alpha component into a color.
     * @param color     hue from which to draw the base rgb values
     * @param alpha     transparency (alpha) value ranging 0-255
     * @return color with alpha value
     * @throws IllegalArgumentException if alpha value is 
     * outside valid bounds (0-255)
     */
	public static Color setAlpha(Color color, int alpha) {
		if (alpha > 255 || alpha < 0) throw new IllegalArgumentException();
		int rgb = color.getRGB();
		rgb &= (1 << 24) - 1; // Blanks alpha value
		rgb |= alpha << 24; // Resets it to new alpha
		return new Color(rgb, true);
	}
	
    /**
     * Returns a color with the opposite rgb values of a given color, 
     * retaining its alpha value.
     * @param color     hue from which to determine negative
     * @return color with opposite rgb values
     */
	public static Color getNegative(Color color) {
		int r = 255 - color.getRed();
		int g = 255 - color.getGreen();
		int b = 255 - color.getBlue();
		int a = color.getAlpha();
		return new Color(r, g, b, a);
	}
	
    /**
     * Returns a blend of two colors, leaning a given degree 
     * toward the first.
     * @param first     color to be blended
     * @param second     color to be blended
     * @param perPrim     percentage leanancy toward first color
     * @param shiftAlpha     if true alpha values are included in the 
     * blend, excluded otherwise
     * @throws IllegalArgumentException if perPrim is greater than 
     * one or less than zero
     */
	// Uses: c2 + ((c1 - c2) * per) = (1-per) c2 + per * c1
	public static Color getGradient(Color first, Color second, double perPrim, boolean shiftAlpha) {
		if (perPrim > 1 || perPrim < 0) throw new IllegalArgumentException();
		int red = (int) (perPrim * first.getRed() + (1 - perPrim) * second.getRed());
		int green = (int) (perPrim * first.getGreen() + (1 - perPrim) * second.getGreen());
		int blue = (int) (perPrim * first.getBlue() + (1 - perPrim) * second.getBlue());
		int alpha = shiftAlpha ? (int) (perPrim * first.getAlpha() + (1 - perPrim) * second.getAlpha()) :
			first.getAlpha();
		return new Color(red, green, blue, alpha);
	}
	
    /**
     * Sets the alpha value for all pixels in an image. Any 
     * fully transparent locations are left that way.
     * @param image     picture to be processed
     * @param alpha     alpha value for pixels
     * @throws IllegalArgumentException if alpha value is 
     * outside valid bounds (0-255)
     */
	public static void setAlpha(BufferedImage image, int alpha) {
		if (alpha > 255 || alpha < 0) throw new IllegalArgumentException();
		int width = image.getWidth();
		int height = image.getHeight();
		
		for (int row=0; row < width; ++row) {
			for (int col=0; col < height; ++col) {
				Color original = new Color(image.getRGB(row, col), true);
				if (original.getAlpha() != 0) {
					Color shifted = setAlpha(original, alpha);
					image.setRGB(row, col, shifted.getRGB());
				}
            }
        }
	}
	
    /**
     * Replaces all instances of one color with another.
     * @param image     picture to be processed
     * @param target     color to be replaced
     * @param replacement     replacement color to substitue
     */
    public static void replaceColor(BufferedImage image, Color target, Color replacement) {
		for (Point location : getLocations(image, target)) {
			image.setRGB(location.x, location.y, replacement.getRGB());
		}
	}
    
	// Returns a hash set of all positions with a given color
	private static HashSet <Point> getLocations(BufferedImage buffered, Color color) {
		HashSet <Point> target = new HashSet <Point> ();
		int width = buffered.getWidth();
		int height = buffered.getHeight();
		
		for (int row=0; row < width; ++row) {
			for (int col=0; col < height; ++col) {
				if (buffered.getRGB(row, col) == color.getRGB()) {
					target.add(new Point(row, col));
				}
			}
		}
		return target;
	}
	
    /**
     * Shifts the hue (rgb values) of the image toward a 
     * certain color. This can be useful for lightening, 
     * darkening, as well as simply recoloing an image.
     * @param image     picture to be processed
     * @param shift     color to which to shift the hue
     * @param percentage     percentage to which to shift the rgb values
     * @param shiftAlpha     if true the alpha values are alterd
     * @throws IllegalArgumentException if percentage is greater than 
     * one or less than zero
     */
    public static void tint(BufferedImage image, Color shift, double percentage, boolean shiftAlpha) {
		if (percentage > 1 || percentage < 0) throw new IllegalArgumentException();
		int width = image.getWidth();
		int height = image.getHeight();
		
		for (int row=0; row < width; ++row) {
			for (int col=0; col < height; ++col) {
				Color original = new Color(image.getRGB(row, col), true);
				Color blend = getGradient(original, shift, 1.0 - percentage, shiftAlpha);
				image.setRGB(row, col, blend.getRGB());
            }
        }
    }
    
    /**
     * Removes all color from an image. Original version written by
     * Marty Stepp.
     * @param image     picture to be processed
     */
	public static void grayscale(BufferedImage image) {
		int width = image.getWidth();
		int height = image.getHeight();

		for (int row=0; row < width; ++row) {
			for (int col=0; col < height; ++col) {
                int c = image.getRGB(row, col);
                int r = (((c >> 16) & 0xff) + ((c >> 8) & 0xff) + (c & 0xff)) / 3;
                image.setRGB(row, col, (0xff << 24) | (r << 16) | (r << 8) | r);
            }
        }
	}
	
    /**
     * Replaces all instances of a color when it borders a different 
     * color. This does not consider diagnal positions.
     * @param image     picture to be processed
     * @param target     color to be replaced
     * @param replacement     color of outline
     * @param thickness     width of border by repeated iterations
     */
	public static void outline(BufferedImage image, Color target, Color replacement, int thickness) {
		outlineGradient(image, target, replacement, replacement, thickness);
	}
	
    /**
     * Replaces all instances of a color when it borders a different 
     * color. The fill is a gradient, gradually going from the first 
     * replacement to the second over the thickness of the border. This 
     * does not consider diagnal positions.
     * @param image     picture to be processed
     * @param target     color to be replaced
     * @param firstRep     initial color of outline
     * @param secondRep     ending color of outline
     * @param thickness     width of border by repeated iterations
     * @throws IllegalArgumentException if thickness is less than zero
     */
	public static void outlineGradient(BufferedImage image, Color target, Color firstRep, Color secondRep, int thickness) {
		if (thickness < 0) throw new IllegalArgumentException();
		double increment = thickness > 1 ? 1/(1.0*thickness-1) : 0;
		
		// Constructs initial set of positions bordering target color
		HashSet <Point> targetPositions = new HashSet <Point> ();
		int width = image.getWidth();
		int height = image.getHeight();
		for (int row=0; row < width; ++row) {
			for (int col=0; col < height; ++col) {
				if (image.getRGB(row, col) == target.getRGB()) {
					Point pt = new Point(row, col);
					for (Direction dir : Direction.values()) {
						Point border = dir.getOffset(pt);
						if (!inBounds(border, image)) continue;
						if (image.getRGB(border.x, border.y) != target.getRGB()) {
							targetPositions.add(border);
						}
					}
				}
			}
		}
		
		for (int i=0; i < thickness; ++i) {
			Color fill = getGradient(firstRep, secondRep, 1.0 - i*increment, true);
			targetPositions = outline(image, targetPositions, target, fill);
		}
	}
	
	// Uses positions that border target color to fill target and expand border
	private static HashSet <Point> outline(BufferedImage buffered, HashSet <Point> border, Color target, Color replacement) {
		HashSet <Point> newBorder = new HashSet <Point> ();
		for (Point loc : border) {
			for (Direction dir : Direction.values()) {
				Point tmp = dir.getOffset(loc);
				if (!inBounds(tmp, buffered)) continue;
				if (buffered.getRGB(tmp.x, tmp.y) == target.getRGB()) {
					newBorder.add(tmp);
				}
			}
		}
		
		for (Point loc : newBorder) {
			buffered.setRGB(loc.x, loc.y, replacement.getRGB());
		}
		
		return newBorder;
	}
	
	/* Old code for outline filter (new filter runs approximately seven times faster)
	public static void outlineGradient(BufferedImage image, Color target, Color firstRep, Color secondRep, int thickness) {
		if (thickness < 0) throw new IllegalArgumentException();
		double increment = thickness > 1 ? 1/(1.0*thickness-1) : 0;
		HashSet <Point> targetPositions = getLocations(image, target);
		
		for (int i=0; i < thickness; ++i) {
			HashSet <Point> border = getOutline(image, targetPositions, target);
			Color fill = getGradient(firstRep, secondRep, 1.0 - i*increment, true);
			if (fill.equals(target)) {
				double offset = i == 0 ? -0.01 : 0.01;
				fill = getGradient(firstRep, secondRep, 1.0 - i*increment + offset, true);
			}
	        for (Point location : border) {
	        	targetPositions.remove(location);
				image.setRGB(location.x, location.y, fill.getRGB());
			}
		}
	}
	
	// Returns a subset of points that border a different color
	private static HashSet <Point> getOutline(BufferedImage image, HashSet <Point> positions, Color target) {
		HashSet <Point> outline = new HashSet <Point> ();
		int width = image.getWidth();
		int height = image.getHeight();
		int targetRGB = target.getRGB();
		
		for (Point location : positions) {
			int x = location.x;
			int y = location.y;
			boolean borders = false;
			borders |= x-1 >= 0 && image.getRGB(x-1, y) != targetRGB;
			borders |= x+1 < width && image.getRGB(x+1, y) != targetRGB;
			borders |= y-1 >= 0 && image.getRGB(x, y-1) != targetRGB;
			borders |= y+1 < height && image.getRGB(x, y+1) != targetRGB;
			if (borders && image.getRGB(x, y) == targetRGB) {
				outline.add(new Point(x, y));
			}
		}
        return outline;
	}
	*/
	
    /**
     * Create a mirror image along an axis. Original version written by
     * Marty Stepp.
     * @param image     picture to be processed
     * @param vertical     swaps top/bottom if true, left/right otherwise
     */
	public static void flip(BufferedImage image, boolean vertical) {
		int width = image.getWidth();
		int height = image.getHeight();

		if (vertical) {
	        for (int row=0; row < height / 2; ++row) {
	            for (int col=0; col < width; ++col) {
					Point p1 = new Point(col, row);
					Point p2 = new Point(col, height - row - 1);
	                swap(image, p1, p2);
	            }
	        }
		} else {
	        for (int row = 0; row < height; row++) {
	            for (int col = 0; col < width / 2; col++) {
					Point p1 = new Point(col, row);
					Point p2 = new Point(width - col - 1, row);
	                swap(image, p1, p2);
	            }
	        }
		}
	}
	
    /**
     * Create an instance of the image rotated ninety degrees.
     * @param image     picture to be processed
     * @param clockwise     rotates clockwise if true, counterclockwise otherwise
     * @return rotated instance of the image
     */
	public static BufferedImage rotate(BufferedImage image, boolean clockwise) {
		int width = image.getWidth();
		int height = image.getHeight();
		BufferedImage rotated = new BufferedImage(height, width, BufferedImage.TYPE_INT_ARGB);

		for (int row=0; row < width; ++row) {
			for (int col=0; col < height; ++col) {
				int x, y;
				int color = image.getRGB(row, col);

				if (clockwise) {
					x = height - col - 1;
					y = row;
				} else {
					x = col;
					y = width - row - 1;
				}
				rotated.setRGB(x, y, color);
			}
		}
		return rotated;
	}
	
	// Switches the color of two pixels
	private static void swap(BufferedImage image, Point p1, Point p2) {
		int color1 = image.getRGB(p1.x, p1.y);
		int color2 = image.getRGB(p2.x, p2.y);
		image.setRGB(p1.x, p1.y, color2);
		image.setRGB(p2.x, p2.y, color1);
	}
}
