package util.media;
/** Stamp class
 * @author Damian Johnson
 * @version Septmber 8, 2006
 */

import java.awt.*;
import util.*;

/** Simple and encapsulized means of constructing and 
 * adding predefined geometric figures to displays. 
 * Properties are configured, then the stamp can be 
 * used to add the figure to specific coordinates.
 * 
 * Some shapes respect an orientation component which 
 * dictate which direction they point. This is defaulted 
 * to Direction.UP if a direction, Corner.UPPER_LEFT if 
 * a corner, and null otherwise. The arguements used by 
 * shapes are:
 * Shape.TRIANGLE -> Direction or Corner (defaults to 
 *      Direction.UP)
 * Shape.RIGHT_TRIANGLE -> Corner
 * Shape.HEXAGON -> Direction
 * Shape.CORNER -> Corner
 */
public class Stamp implements Cloneable {
	private Shape shape;
	private Paint paint;
	private Orientation gradientDir = null; // If set, a gradient is used
	private Dimension dimensions;
	private double scale = 1;
	private Orientation orientation;
	
    /**
     * Constructor for a stamp with square dimensions. 
     * @param shape     shape of figure
     * @param paint     paint of figure
     * @param sides     width and height of figure
     */
	public Stamp(Shape shape, Paint paint, int sides) {
		this(shape, paint, new Dimension(sides, sides));
	}
	
    /**
     * Constructs a stamp with defined dimensions.
     * @param shape     shape of figure
     * @param paint     paint of figure
     * @param dimension     size of figure
     */
	public Stamp(Shape shape, Paint paint, Dimension dimension) {
		setShape(shape);
		this.paint = paint;
		this.dimensions = dimension;
	}
	
    /**
     * Returns the shape of the figure.
     * @return shape of figure
     */
	public Shape getShape() {
		return this.shape;
	}
	
    /**
     * Sets the shape of the figure. The orientation 
     * field is set to the shape's default orientation.
     * @param shape     shape of figure
     */
	public void setShape(Shape shape) {
		this.shape = shape;
		if (this.shape.isLegalOrientation(Direction.UP)) {
			this.orientation = Direction.UP;
		} else if (this.shape.isLegalOrientation(Corner.UPPER_LEFT)) {
			this.orientation = Corner.UPPER_LEFT;
		} else {
			this.orientation = null;
		}
	}
	
    /**
     * Returns the paint of the figure. If a 
     * localized gradient is set then this 
     * returns an instance of a GradientPaint 
     * which can be used to determine the 
     * colors used in the gradient.
     * @return paint of figure
     */
	public Paint getPaint() {
		return this.paint;
	}
	
    /**
     * Sets the paint of the figure.
     * @param paint     paint of figure
     */
	public void setPaint(Paint paint) {
		this.paint = paint;
		this.gradientDir = null;
	}
	
    /**
     * Returns the dimensions of the figure.
     * @return dimensions of figure
     */
	public Dimension getDimensions() {
		return this.dimensions;
	}
	
    /**
     * Sets the dimensions of the figure.
     * @param dimension     size of figure
     */
	public void setDimensions(Dimension dimension) {
		this.dimensions = dimension;
	}
	
    /**
     * Returns the percent of the dimensions occupied 
     * by the figure.
     * @return percentage scale of figure
     */
	public double getScale() {
		return this.scale;
	}
	
    /**
     * Sets the percent of the dimensions occupied 
     * by the figure.
     * @param scale     percentage scale of figure
     * @throws IllegalArgumentException if scale is 
     * greater than one or less than zero.
     */
	public void setScale(double scale) {
		if (scale > 1 || scale < 0) throw new IllegalArgumentException();
		this.scale = scale;
	}
	
    /**
     * Returns the orientation of the figure.
     * @return orientation of figure
     */
	public Orientation getOrientation() {
		return this.orientation;
	}
	
    /**
     * Sets the orientation used by the figure.
     * @param orientation     top facing direction of pattern
     * @throws IllegalArgumentException if the orientation is 
     * incompatable with the shape
     */
	public void setOrientation(Orientation orientation) {
		if (!this.shape.isLegalOrientation(orientation)) 
			throw new IllegalArgumentException();
		this.orientation = orientation;
	}
	
    /**
     * Returns if the stamp is set to use its localized 
     * gradient or not.
     * @return true if stamp uses a localized gradient, false 
     * otherwise
     */
	public boolean isUsingGradient() {
		return this.gradientDir != null;
	}
	
    /**
     * Returns the orientation of the localized gradient 
     * (null if none is set).
     * @return orientation direction of localized gradient
     */
	public Orientation getGradientOrientation() {
		return this.gradientDir;
	}
	
    /**
     * Sets the paint to a gradient localized within 
     * the dimensions of the stamp.
     * @param first     color in gradient
     * @param second     color in gradient
     * @param orientation     direction of gradient
     */
	public void setGradient(Color first, Color second, Orientation orientation) {
		this.paint = new GradientPaint(new Point(0, 0), first, new Point(0, 0), second);
		this.gradientDir = orientation;
	}
	
    /**
     * Draws the figure centered within its dimensions.
     * @param g2     graphic object in which to draw figure
     * @param anchor     upper left coordinates of dimension
     * @param fill     if true the instance is filled, 
     * otherwise only an outline is drawn
     */
	public void draw(Graphics g, Point anchor, boolean fill) {
		Graphics2D g2 = (Graphics2D) g;
		Paint previous = g2.getPaint();
		if (this.gradientDir != null) {
			Point firstPt = getSide(anchor, this.gradientDir.reverse());
			Point secondPt = getSide(anchor, this.gradientDir);
			Color firstColor = ((GradientPaint) this.paint).getColor1();
			Color secondColor = ((GradientPaint) this.paint).getColor2();
			g2.setPaint(new GradientPaint(firstPt, firstColor, secondPt, secondColor));
		} else {
			g2.setPaint(this.paint);
		}
		
		int x = anchor.x + (int) Math.ceil(this.dimensions.width * (1 - this.scale) / 2);
		int y = anchor.y + (int) Math.ceil(this.dimensions.height * (1 - this.scale) / 2);
		int width = (int) Math.floor(this.dimensions.width * this.scale);
		int height = (int) Math.floor(this.dimensions.height * this.scale);
		int[] xArray, yArray;
		
		if (this.shape == Shape.OVAL) {
			if (fill) g2.fillOval(x, y, width, height);
			else g2.drawOval(x, y, width, height);
		} else if (this.shape == Shape.TRIANGLE) {
			if (this.orientation == Direction.UP) {
				xArray = new int[] {x, x + width, x + width / 2};
				yArray = new int[] {y + height, y + height, y};
			} else if (this.orientation == Direction.RIGHT) {
				xArray = new int[] {x, x, x + width};
				yArray = new int[] {y, y + height, y + height / 2};
			} else if (this.orientation == Direction.DOWN) {
				xArray = new int[] {x, x + width, x + width / 2};
				yArray = new int[] {y, y, y + height};
			} else if (this.orientation == Direction.LEFT){
				xArray = new int[] {x + width, x + width, x};
				yArray = new int[] {y, y + height, y + height / 2};
			} else if (this.orientation == Corner.UPPER_LEFT){
				xArray = new int[] {x, x + width, x + width / 2};
				yArray = new int[] {y, y + height / 2, y + height};
			} else if (this.orientation == Corner.UPPER_RIGHT){
				xArray = new int[] {x + width, x + width / 2, x};
				yArray = new int[] {y, y + height, y + height / 2};
			} else if (this.orientation == Corner.LOWER_RIGHT){
				xArray = new int[] {x + width, x, x + width / 2};
				yArray = new int[] {y + height, y + height / 2, y};
			} else {
				xArray = new int[] {x, x + width / 2, x + width};
				yArray = new int[] {y + height, y, y + height / 2};
			}
			if (fill) g2.fillPolygon(xArray, yArray, 3);
			else g2.drawPolygon(xArray, yArray, 3);
		} else if (this.shape == Shape.RIGHT_TRIANGLE) {
			if (this.orientation == Corner.UPPER_LEFT) {
				xArray = new int[] {x, x + width, x};
				yArray = new int[] {y, y, y + height};
			} else if (this.orientation == Corner.UPPER_RIGHT) {
				xArray = new int[] {x + width, x + width, x};
				yArray = new int[] {y, y + height, y};
			} else if (this.orientation == Corner.LOWER_RIGHT) {
				xArray = new int[] {x + width, x, x + width};
				yArray = new int[] {y + height, y + height, y};
			} else {
				xArray = new int[] {x, x, x + width};
				yArray = new int[] {y + height, y, y + height};
			}
			if (fill) g2.fillPolygon(xArray, yArray, 3);
			else g2.drawPolygon(xArray, yArray, 3);
		} else if (this.shape == Shape.RECTANGLE) {
			if (fill) g2.fillRect(x, y, width, height);
			else g2.drawRect(x, y, width, height);
		} else if (this.shape == Shape.DIAMOND) {
			xArray = new int[] {x, x + width / 2, x + width, x + width / 2};
			yArray = new int[] {y + height / 2, y, y + height / 2, y + height};
			if (fill) g2.fillPolygon(xArray, yArray, 4);
			else g2.drawPolygon(xArray, yArray, 4);
		} else if (this.shape == Shape.HEXAGON) {
			if (this.orientation == Direction.UP || this.orientation == Direction.DOWN) {
				xArray = new int[] {x, x, x + width / 2, x + width, x + width, x + width / 2};
				yArray = new int[] {y + height * 2 / 3, y + height / 3, y, y + height / 3, y + height * 2 / 3, y + height};
			} else {
				xArray = new int[] {x + width * 2 / 3, x + width / 3, x, x + width / 3, x + width * 2 / 3, x + width};
				yArray = new int[] {y, y, y + height / 2, y + height, y + height, y + height / 2};
			}
			if (fill) g2.fillPolygon(xArray, yArray, 6);
			else g2.drawPolygon(xArray, yArray, 6);
		} else if (this.shape == Shape.CORNER) {
			Corner cor = (Corner) this.orientation;
			if (cor.isUpper()) g2.drawLine(x, y, x + width - 1, y);
			else g2.drawLine(x, y + height - 1, x + width - 1, y + height - 1);
			
			if (cor.isLeft()) g2.drawLine(x, y, x, y + height - 1);
			else g2.drawLine(x + width - 1, y, x + width - 1, y + height - 1);
		}
		g2.setPaint(previous);
	}
	
    /**
     * Draws smaller copies of the stamp on the sides of the 
     * perimeter. If the shape accepts Direction orientations then 
     * the orientation rotates clockwise with each successive 
     * corner, starting with the top.
     * @param g     graphic object in which to border
     * @param anchor     upper left coordinates
     * @param fill     if true the instance is filled, 
     * otherwise only an outline is drawn
     */
	public void borderSides(Graphics g, Point anchor, boolean fill) {
		int width = (int) Math.floor(this.dimensions.getWidth() * this.scale / 2);
		if (width % 2 == 1) --width;
		int height = (int) Math.floor(this.dimensions.getHeight() * this.scale / 2);
		if (height % 2 == 1) --height;
		Stamp sideStamp = new Stamp(this.shape, this.paint, new Dimension(width, height));
		sideStamp.setOrientation(this.orientation);
		if (this.gradientDir != null) {
			Color firstColor = ((GradientPaint) this.paint).getColor1();
			Color secondColor = ((GradientPaint) this.paint).getColor2();
			sideStamp.setGradient(firstColor, secondColor, this.gradientDir);
		}
		
		for (Direction dir : Direction.values()) {
			int x = anchor.x;
			if (dir.isVertical()) {
				x += (this.dimensions.width - width) / 2;
			} else if (dir == Direction.RIGHT) {
				x += this.dimensions.width - width;
			}
			
			int y = anchor.y;
			if (!dir.isVertical()) {
				y += (this.dimensions.height - height) / 2;
			} else if (dir == Direction.DOWN) {
				y += this.dimensions.height - height;
			}
			sideStamp.draw(g, new Point(x, y), fill);
			
			Orientation orientation = sideStamp.getOrientation();
			if (orientation instanceof Direction) {
				sideStamp.setOrientation(orientation.right());
			}
		}
	}
	
    /**
     * Draws smaller copies of the stamp in the corners of the 
     * perimeter. If the shape accepts Corner orientations then 
     * the orientation rotates clockwise with each successive 
     * corner, starting with the upper left.
     * @param g     graphic object in which to border
     * @param anchor     upper left coordinates
     * @param fill     if true the instance is filled, 
     * otherwise only an outline is drawn
     */
	public void borderCorners(Graphics g, Point anchor, boolean fill) {
		int width = (int) Math.floor(this.dimensions.getWidth() * this.scale / 2);
		int height = (int) Math.floor(this.dimensions.getHeight() * this.scale / 2);
		Stamp cornerStamp = new Stamp(this.shape, this.paint, new Dimension(width, height));
		cornerStamp.setOrientation(this.orientation);
		if (this.gradientDir != null) {
			Color firstColor = ((GradientPaint) this.paint).getColor1();
			Color secondColor = ((GradientPaint) this.paint).getColor2();
			cornerStamp.setGradient(firstColor, secondColor, this.gradientDir);
		}
		
		for (Corner cor : Corner.values()) {
			int x = cor.isLeft() ? anchor.x : anchor.x + this.dimensions.width - width;
			int y = cor.isUpper() ? anchor.y : anchor.y + this.dimensions.height - height;
			cornerStamp.draw(g, new Point(x, y), fill);
			
			Orientation orientation = cornerStamp.getOrientation();
			if (orientation instanceof Corner) {
				cornerStamp.setOrientation(orientation.right());
			}
		}
	}
	
    /**
     * Constructs a deep copy.
     * @return clone of stamp
     */
	public Stamp clone() {
		Stamp clone = new Stamp(this.shape, this.paint, this.dimensions);
		clone.setScale(getScale());
		clone.setOrientation(getOrientation());
		if (isUsingGradient()) {
			Color c1 = ((GradientPaint) this.paint).getColor1();
			Color c2 = ((GradientPaint) this.paint).getColor2();
			Orientation orientation = getGradientOrientation();
			clone.setGradient(c1, c2, orientation);
		}
		return clone;
	}
	
	// Returns the point within the dimensions of a given side
	private Point getSide(Point loc, Orientation orientation) {
		int width = this.dimensions.width;
		int height = this.dimensions.height;
		if (orientation instanceof Direction) {
			switch ((Direction) orientation) {
			case UP: return new Point(loc.x + width / 2, loc.y);
			case RIGHT: return new Point(loc.x + width, loc.y + height / 2);
			case DOWN: return new Point(loc.x + width / 2, loc.y + height);
			case LEFT: return new Point(loc.x, loc.y + height / 2);
			}
		} else {
			switch ((Corner) orientation) {
			case UPPER_LEFT: return new Point(loc.x, loc.y);
			case UPPER_RIGHT: return new Point(loc.x + width, loc.y);
			case LOWER_RIGHT: return new Point(loc.x + width, loc.y + height);
			case LOWER_LEFT: return new Point(loc.x, loc.y + height);
			}
		}
		return null;
	}
	
	/** Representation of a variety of shapes.
	 */
	public enum Shape {
		OVAL, TRIANGLE, RIGHT_TRIANGLE, RECTANGLE, DIAMOND, 
		HEXAGON, CORNER;
		
	    /**
	     * Returns the shape representation of a string. This 
	     * is not case sensative.
	     * @param str     string representation of a shape
	     * @return shape associated with a string
	     */
	    public static Shape fromString(String str) {
	    	if (str == null) throw new NullPointerException();
	    	else if (str.equals("Right Triangle")) return RIGHT_TRIANGLE;
	    	str = str.toUpperCase();
	    	try {
	        	return valueOf(str);
	    	} catch (IllegalArgumentException exc) {
	    		return null;
	    	}
	    }
	    
	    /**
	     * Returns if a given arguement would be acceptable 
	     * if set as the orientation.
	     * @param arg     arguement to be tested
	     * @return arguement type of orientation
	     */
		public boolean isLegalOrientation(Orientation arg) {
			switch (this) {
			case TRIANGLE: return true;
			case RIGHT_TRIANGLE: return arg instanceof Corner;
			case HEXAGON: return arg instanceof Direction;
			case CORNER: return arg instanceof Corner;
			default: return arg == null;
			}
		}
		
	    /**
	     * Returns a string representation of this shape.
	     * @return string representation of this enum
	     */
		public String toString() {
			switch (this) {
			case OVAL: return "Oval";
			case TRIANGLE: return "Triangle";
			case RIGHT_TRIANGLE: return "Right Triangle";
			case RECTANGLE: return "Rectangle";
			case DIAMOND: return "Diamond";
			case HEXAGON: return "Hexagon";
			case CORNER: return "Corner";
			default: return null;
			}
		}
	}
}
