package util.squareGrid;
/** Board class
 * @author Damian Johnson
 * @version August 27, 2006
 */

import java.awt.Point;
import java.util.*;
import util.Direction;

/** Representation of a board consisting of square tiles. This 
 * provides useful utilities for positions relative to the dimensions. 
 * Contents are observed and exceptions thrown if they overlap or 
 * fall out of bounds.
 */
public class Board {
	private int width, height;
	private boolean allowOverlap;
	private PositionChecker positionChecker = new PositionChecker();
	private HashSet <Placeable> contents = new HashSet <Placeable> ();
	
    /**
     * Constructor of a board of specific dimensions that throws 
     * exceptions for contents with overlapping positions.
     * @param width     board width
     * @param height     board height
     */
	public Board(int width, int height) {
		this(width, height, false);
	}
	
    /**
     * Constructor of a board of specific dimensions.
     * @param width     board width
     * @param height     board height
     * @param allowOverlap     if true conflicting positions are 
     * tolerated, otherwise exceptions are thrown
     */
	public Board(int width, int height, boolean allowOverlap) {
		this.width = width;
		this.height = height;
		this.allowOverlap = allowOverlap;
	}
	
    /**
     * Returns the width dimension of the board.
     * @return width of the board
     */
	public int getWidth() {
		return this.width;
	}
	
    /**
     * Returns the height dimension of the board.
     * @return height of the board
     */
	public int getHeight() {
		return this.height;
	}
	
    /**
     * Returns current placeable contents of the board.
     * @return set of registered placeables
     */
	public HashSet <Placeable> getContents() {
		return new HashSet <Placeable> (this.contents);
	}
	
    /**
     * Returns if the board allows overlap or not.
     * @return true if conflicing positions are tolerated, false 
     * if exceptions are thrown
     */
	public boolean allowsOverlap() {
		return this.allowOverlap;
	}
	
    /**
     * Produces the contents of a given coorindate (null if empty).
     * @param location     position of unit to be obtained
     * @return placeable object at locaiton, null if empty
     */
	public Placeable getLocation(Point location) {
		for (Placeable placeable : this.contents) {
			if (placeable.contains(location)) return placeable;
		}
		return null;
	}
	
    /**
     * Returns the edge in a given direction from a point.
     * @param location     position to be referenced
     * @param dir     direction of the edge to be returned
     * @return point on the given edge of the board
     * @throws IllegalArgumentException location is outside 
     * the board's dimensions
     */
	public Point getEdge(Point location, Direction dir) {
		if (!inBounds(location)) throw new IllegalArgumentException();
		switch (dir) {
		case UP: return new Point(location.x, 1);
		case RIGHT: return new Point(this.width, location.y);
		case DOWN: return new Point(location.x, this.height);
		case LEFT: return new Point(1, location.y);
		default: return null;
		}
	}
	
    /**
     * Determines if a position is within the borders of the board.
     * @param location     position to be checked
     * @return true if location is within the dimensions of the 
     * board, false otherwise
     */
	public boolean inBounds(Point location) {
		boolean inBounds = true;
		inBounds &= location.x < this.width;
		inBounds &= location.x >= 0;
		inBounds &= location.y < this.height;
		inBounds &= location.y >= 0;
		return inBounds;
	}
	
    /**
     * Determines if all points in a collection are within bounds.
     * @param locations     collection of positions to be checked
     * @return true if all positoins are within the board's 
     * dimensions, false otherwise
     */
	public boolean inBounds(Collection <Point> locations) {
		for (Point loc : locations) {
			if (!inBounds(loc)) return false;
		}
		return true;
	}
	
    /**
     * Determines if a position lacks a pressence.
     * @param location     position to be checked
     * @return true if the contents does not include the location, 
     * false otherwise
     */
	public boolean isEmpty(Point location) {
		return getLocation(location) == null;
	}
	
    /**
     * Determines if all points in a collection lack a pressence.
     * @param locations     collection of positions to be checked
     * @return true if all positoins are empty, false otherwise
     */
	public boolean isEmpty(Collection <Point> locations) {
		for (Point location : locations) {
			if (!isEmpty(location)) return false;
		}
		return true;
	}
	
    /**
     * Adds a placeable to the contents of the board, tracking 
     * changes in its positions for overlap and bounds.
     * @param placable     placeable object to be added
     * @throws BoundsException if placeable's out of bounds
     * @throws CollisionException if a position is occupied 
     * and overlap is not allowed
     */
	public void add(Placeable placeable) {
		if (!inBounds(placeable.getLocations())) {
			Point anchor = placeable.getAnchor();
			throw new BoundsException(this, placeable, anchor);
		} else if (!this.allowOverlap && !isEmpty(placeable.getLocations())) {
			Point anchor = placeable.getAnchor();
			throw new CollisionException(this, placeable, anchor);
		} else {
			placeable.getMovementEvent().add(this.positionChecker);
			this.contents.add(placeable);
		}
	}
	
    /**
     * Removes a placeable from the contents and consideration 
     * within the board.
     * @param placable     placeable object to be removed
     * @return true if the placable is successfully removed, 
     * false otherwise.
     */
	public boolean remove(Placeable placeable) {
		if (this.contents.contains(placeable)) {
			placeable.getMovementEvent().remove(this.positionChecker);
			return this.contents.remove(placeable);
		} else return false;
	}
	
    /**
     * Removes all placeables from the contents and and 
     * consideration within the board.
     */
	public void clear() {
		while (!this.contents.isEmpty()) {
			remove(this.contents.iterator().next());
		}
	}
	
	// Fired before board contents change positions, throwing 
	// BoundsExceptions and CollisonExceptions if applicable
	private class PositionChecker implements Observer {
		public void update(Observable o, Object arg) {
			if (arg instanceof Placeable.PlacementData) {
				Placeable mover = ((Placeable.PlacementData) arg).mover;
				Point anchor = ((Placeable.PlacementData) arg).attemptedAnchor;
				Board board = Board.this;
				
				for (Point loc : PointTools.shift(mover.getLocations(), anchor)) {
					if (!inBounds(loc)) {
						throw new BoundsException(board, mover, anchor);
					}
				}
				
				if (!allowOverlap) {
					for (Point loc : PointTools.shift(mover.getLocations(), anchor)) {
						Placeable occupant = getLocation(loc);
						if (occupant != null && occupant != mover) {
							throw new CollisionException(board, mover, anchor);
						}
					}
				}
			}
		}
	}
}
