package util.squareGrid;
/** Placeable class
 * @author Damian Johnson
 * @version September 4, 2006
 */
// Possibly useful:
// canMove(Direction), canSetAnchor(Point), canFace(Direction)

import java.awt.*;
import java.awt.image.*;
import java.util.*;

import util.*;
import util.Event;
import util.media.GraphicTools;

/** Representation of an object that can be placed on a grid using 
 * Cartesian coordinates.
 */
public class Placeable implements Cloneable, Iterable <Point> {
	private HashMap <Direction, HashSet <Point>> patterns;
	private HashMap <Direction, BufferedImage> images;
	private Point anchor;
	private Direction orientation = Direction.UP;
	private Event movementEvent = new Event();
	
    /**
     * Constructor for a single location, initially facing upwards.
     * @param location     inital location
     */
	public Placeable(Point location) {
		this(PointTools.wrapper(location));
	}
	
    /**
     * Constructor for a set of locations, initially facing upwards.
     * @param locations     inital locations
     * @throws IllegalArgumentException if pattern is empty
     */
	public Placeable(Set <Point> locations) {
    	if (locations.isEmpty()) throw new IllegalArgumentException();
    	this.patterns = new HashMap <Direction, HashSet <Point>> ();
    	this.images = new HashMap <Direction, BufferedImage> ();
		this.anchor = PointTools.getAnchor(locations);
		
		Direction dir = Direction.UP;
		HashSet <Point> set = new HashSet <Point> (locations);
		for (int i=0; i < Direction.values().length; ++i) {
			set = PointTools.shift(set, new Point(0, 0));
			this.patterns.put(dir, set);
			Point pivot = PointTools.getPivot(set, dir);
			set = new HashSet <Point> (PointTools.rotate(set, true, pivot));
			dir = dir.right();
		}
	}
	
    /**
     * Returns the orientation.
     * @return current orientation
     */
	public Direction getDirection() {
		return this.orientation;
	}
	
    /**
     * Returns a clone of the points that this occupies.
     * @return locations occupied
     */
	public HashSet <Point> getLocations() {
		HashSet <Point> pattern = this.patterns.get(this.orientation);
		return PointTools.shift(pattern, this.anchor);
	}
	
    /**
     * Returns the orriginal pattern from which the placeable 
     * was constrcted. This is the upward facing figure anchored 
     * to the zero coordinates.
     * @return placeable's layout
     */
	public HashSet <Point> getPattern() {
		HashSet <Point> pattern = this.patterns.get(Direction.UP);
		return PointTools.copy(pattern);
	}
	
    /**
     * Returns the upper left point of the border, often useful 
     * for image locationing and reference.
     * @return upper left location
     */
	public Point getAnchor() {
		return this.anchor;
	}
	
    /**
     * Returns the image representation for the current orientation 
     * if an image has been set.
     * @return image representation of placeable
     */
	public BufferedImage getImage() {
		return getImage(getDirection());
	}
	
    /**
     * Returns the image representation for a given orientation if 
     * an image has been set.
     * @param dir     image orientation
     * @return image representation of placeable
     */
	public BufferedImage getImage(Direction dir) {
		if (this.images.isEmpty()) return null;
		else return this.images.get(dir);
	}
	
    /**
     * Returns the image for the current orientation stretched or 
     * shrank to fit a specific scaling. The resulting dimensions 
     * of the image is the scale times the placeable's width and 
     * height respectively.
     * @param scale     pixel multiple of dimensions
     * @return scaled image representation of placeable
     */
	public BufferedImage getImage(int scale) {
		if (this.images.isEmpty()) return null;
		else {
			BufferedImage img = getImage();
			int width = getWidth() * scale;
			int height = getHeight() * scale;
			double xScale = (double) width / img.getWidth();
			double yScale = (double) height / img.getHeight();
			
			BufferedImage scaled = new BufferedImage(width, height, BufferedImage.TYPE_INT_ARGB);
			Graphics2D scaledG2 = (Graphics2D) scaled.getGraphics();
			scaledG2.scale(xScale, yScale);
			scaledG2.drawImage(img, 0, 0, null);
			return scaled;
		}
	}
	
    /**
     * Returns the width dimension.
     * @return width of this placeable object
     */
	public int getWidth() {
		return PointTools.getWidth(getLocations());
	}
	
    /**
     * Returns the height dimension.
     * @return height of this placeable object
     */
	public int getHeight() {
		return PointTools.getHeight(getLocations());
	}
	
    /**
     * Converts a point between the placeable's 
     * current coordinates and its pattern.
     * @param location     location to be converted
     * @param toPattern     if true the conversion is 
     * from actual coordinates to pattern, otherwise 
     * the other way
     * @return related point in the pattern or coordinates
     */
	public Point getRelation(Point location, boolean toPattern) {
		if (toPattern && !contains(location)) return null;
		else if (!toPattern && !getPattern().contains(location)) return null;
		Point pivot = PointTools.getPivot(getLocations(), this.orientation);
		Point patternPivot = PointTools.getPivot(this.patterns.get(Direction.UP), Direction.UP);
		int x = toPattern ? patternPivot.x - pivot.x : pivot.x - patternPivot.x;
		int y = toPattern ? patternPivot.y - pivot.y : pivot.y - patternPivot.y;
		if (!toPattern) pivot = patternPivot;
		Direction dir = this.orientation;
		if (!toPattern && !dir.isVertical()) dir = dir.reverse();
		
		if (dir == Direction.UP) {
			x += location.x;
			y += location.y;
		} else if (dir == Direction.RIGHT) {
			x += pivot.x + location.y - pivot.y;
			y += pivot.y - location.x + pivot.x;
		} else if (dir == Direction.DOWN) {
			x += location.x + (pivot.x - location.x) * 2;
			y += location.y + (pivot.y - location.y) * 2;
		} else {
			x += pivot.x - location.y + pivot.y;
			y += pivot.y + location.x - pivot.x;
		}
		return new Point(x, y);
	}
	
    /**
     * Returns if this contains a given location.
     * @param location     location to be compared
     * @return if the location is present in the locations
     */
	public boolean contains(Point location) {
		return getLocations().contains(location);
	}
	
    /**
     * Increments the coordinates for a given direction.
     * @param dir     direction to move the coordinates
     */
	public void move(Direction dir) {
		Point newAnchor = dir.getOffset(this.anchor);
		this.movementEvent.fire(new PlacementData(this, newAnchor));
		this.anchor = newAnchor;
	}
	
    /**
     * Shifts the locations to have a specific upper left location.
     * @param target     location of anchor to which points conform
     */
	public void setAnchor(Point target) {
		this.movementEvent.fire(new PlacementData(this, target));
		this.anchor = target;
	}
	
    /**
     * Sets an image representation, constructing its appearance 
     * for each orientation.
     * @param image     upward facing image representation
     */
	public void setImage(BufferedImage image) {
		this.images.clear();
		if (image == null) return;
		Direction dir = Direction.UP;
		for (int i=0; i < Direction.values().length; ++i) {
			this.images.put(dir, image);
			dir = dir.right();
			image = GraphicTools.rotate(image, true);
		}
	}
	
    // Hacks to improve performance when rapidly creating placeables
    public HashMap <Direction, BufferedImage> getImageSet() {
      return this.images;
    }
    
    public void setImageSet(HashMap <Direction, BufferedImage> imageSet) {
      this.images = imageSet;
    }
    
    /**
     * Rotates the placeable to face a given direction.
     * @param dir     direction for the placeable to face
     */
	public void setDirection(Direction dir) {
		// Determines the anchor offset for the rotation
		Point firstPvt = PointTools.getPivot(this.patterns.get(this.orientation), this.orientation);
		Point secondPvt = PointTools.getPivot(this.patterns.get(dir), dir);
		int xDiff = firstPvt.x - secondPvt.x;
		int yDiff = firstPvt.y - secondPvt.y;
		Point newAnchor = new Point(this.anchor.x + xDiff, this.anchor.y + yDiff);
		
		Direction original = this.orientation;
		try {
			this.orientation = dir;
			this.movementEvent.fire(new PlacementData(this, newAnchor));
			this.anchor = newAnchor;
		} catch (PositionException exc) {
			this.orientation = original;
			throw exc;
		}
	}
	
    /**
     * Constructs a deep copy.
     * @return clone of placeable
     */
	public Placeable clone() {
		HashSet <Point> pattern = this.patterns.get(Direction.UP);
		pattern = PointTools.copy(pattern);
		Placeable copy = new Placeable(pattern);
		copy.anchor = new Point(this.anchor);
		copy.orientation = this.orientation;
		
		if (!this.images.isEmpty()) {
			BufferedImage img = this.images.get(Direction.UP);
			img = GraphicTools.copy(img);
			copy.setImage(img);
		}
		return copy;
	}
	
    /**
     * Provides an iterator over the placeable's locations.
     * @return iterator of locations
     */
	public Iterator <Point> iterator() {
		return getLocations().iterator();
	}
	
    /**
     * Provides an observable event to which movement data is sent.
     * @return event for movement data
     */
	protected Event getMovementEvent() {
		return this.movementEvent;
	}
	
	protected class PlacementData {
		public Placeable mover;
		public Point attemptedAnchor;
		
		private PlacementData(Placeable mover, Point attemptedAnchor) {
			this.mover = mover;
			this.attemptedAnchor = attemptedAnchor;
		}
	}
}
