package util.squareGrid;
/** PointTools class
 * @author Damian Johnson
 * @version September 1, 2006
 */

import java.awt.Point;
import java.util.*;
import util.Direction;

/** Utilities for evaluating and modifying points and 
 * collections of points.
 */
public class PointTools {
    /**
     * Returns the shortest distance between two points.
     * @param p1     location to be evaluated
     * @param p2     location to be evaluated
     * @return shortest distance between the two points
     */
	public static int getDistance(Point p1, Point p2) {
		return getDistance(p1, p2, true);
	}
	
    /**
     * Returns the shortest distance between two points.
     * @param p1     location to be evaluated
     * @param p2     location to be evaluated
     * @param diagnals     distance can include diagnal 
     * movement
     * @return shortest distance between the two points
     */
	public static int getDistance(Point p1, Point p2, boolean diagnals) {
    	int xDifference = Math.abs(p1.x - p2.x);
    	int yDifference = Math.abs(p1.y - p2.y);
    	return diagnals ? Math.max(xDifference, yDifference) :
    		xDifference + yDifference;
	}
	
    /**
     * Returns a set of points with a diamond shaped 
     * pattern (empty if radius is less than one).
     * @param center     central position in the diamond
     * @param radius     distance that the diamond reaches 
     * from the center
     * @return locations forming a diamond
     */
	public static HashSet <Point> getDiamond(Point center, int radius) {
		HashSet <Point> locations = new HashSet <Point> ();
		int x = center.x;
		int y = center.y;
		for (int i=0; i<radius; ++i) {
			for (int k=0; k<radius - i; ++k) {
				locations.add(new Point(x+i, y+k));
				if (k != 0) locations.add(new Point(x+i, y-k));
				if (i != 0) locations.add(new Point(x-i, y+k));
				if (i != 0 && k != 0) locations.add(new Point(x-i, y-k));
			}
		}
		return locations;
	}
	
    /**
     * Returns the points surrounding the present positions 
     * (excluding diagnals).
     * @param locations     locations to be checked
     * @return set of bordering positions
     */
	public static HashSet <Point> getBorder(Collection <Point> locations) {
		HashSet <Point> border = new HashSet <Point> ();
		for (Point loc : locations) {
			for (Direction dir : Direction.values()) {
				Point offset = dir.getOffset(loc);
				if (!locations.contains(offset)) {
					border.add(offset);
				}
			}
		}
		return border;
	}
	
    /**
     * Returns the furthest index to a given direction.
     * @param locations     locations to be checked
     * @param dir     direction of the side
     * @return furthest index in a given direction
     * @throws IllegalArgumentException     if no points 
     * are provided
     */
	public static int getBorder(Collection <Point> locations, Direction dir) {
    	if (locations.isEmpty()) throw new IllegalArgumentException();
		int border = dir == Direction.UP || dir == Direction.LEFT ? 
				Integer.MAX_VALUE : Integer.MIN_VALUE;
		
		for (Point loc : locations) {
			if (dir == Direction.UP) border = Math.min(border, loc.y);
			else if (dir == Direction.DOWN) border = Math.max(border, loc.y);
			else if (dir == Direction.LEFT) border = Math.min(border, loc.x);
			else if (dir == Direction.RIGHT) border = Math.max(border, loc.x);
		}
		return border;
	}
	
    /**
     * Returns either the minimum or maximum value for 
     * either of the dimensions in a collection of points.
     * @param locations     locations to be checked
     * @param xDimension     x dimension is evaluated if 
     * true, y dimension otherwise
     * @param max     maximum value is returned if true, 
     * minimum otherwise
     * @return maximum or minimum of a dimension
     * @throws IllegalArgumentException     if no points 
     * are provided
     */
	public static int getBound(Collection <Point> locations, 
			boolean xDimension, boolean max) {
		if (locations.isEmpty()) throw new IllegalArgumentException();
		int bound = max ? Integer.MIN_VALUE : Integer.MAX_VALUE;
		for (Point loc : locations) {
			int value = xDimension ? loc.x : loc.y;
			if (max) bound = Math.max(bound, value);
			else bound = Math.min(bound, value);
		}
		return bound;
	}
	
    /**
     * Returns the upper left position of the border, useful for 
     * image locationing and reference.
     * @param locations     locations to be checked
     * @return upper left location
     */
	public static Point getAnchor(Collection <Point> locations) {
		int left = getBorder(locations, Direction.LEFT);
		int top = getBorder(locations, Direction.UP);
		return new Point(left, top);
	}
	
    /**
     * Returns the width dimension.
     * @param locations     locations to be checked
     * @return width dimension
     */
	public static int getWidth(Collection <Point> locations) {
		return getBorder(locations, Direction.RIGHT) - getBorder(locations, Direction.LEFT) + 1;
	}
	
    /**
     * Returns the height dimension.
     * @param locations     locations to be checked
     * @return height dimension
     */
	public static int getHeight(Collection <Point> locations) {
		return getBorder(locations, Direction.DOWN) - getBorder(locations, Direction.UP) + 1;
	}
	
    /**
     * Returns central location within a collection of 
     * points for standard rotations.
     * @param locations     locations to be checked
     * @param orientation     forward direction of the 
     * collection (for rounding)
     * @return central location in the collection
     */
	public static Point getPivot(Collection <Point> locations, Direction orientation) {
		int x, y;
		if (orientation == Direction.UP || orientation == Direction.LEFT) {
			x = getBorder(locations, Direction.LEFT) + (getWidth(locations) - 1) / 2;
			y = getBorder(locations, Direction.UP) + (getHeight(locations) - 1) / 2;
		} else {
			x = getBorder(locations, Direction.RIGHT) - (getWidth(locations) - 1) / 2;
			y = getBorder(locations, Direction.DOWN) - (getHeight(locations) - 1) / 2;
		}
		return new Point(x, y);
	}
	
    /**
     * Deep clone of a collection of points.
     * @param locations     locations to be copied
     * @return deep copy of the passed collection
     */
	public static HashSet <Point> copy(Collection <Point> locations) {
		HashSet <Point> copy = new HashSet <Point> ();
		for (Point loc : locations) {
			copy.add(new Point(loc));
		}
		return copy;
	}
	
    /**
     * Constructs a collection wrapper for a single point.
     * @param location     contents of the wrapper
     * @return hash set containing the point
     */
	public static HashSet <Point> wrapper(Point location) {
		return wrapper(new Point[] {location});
	}
	
    /**
     * Constructs a collection wrapper for an array of points.
     * @param locations     contents of the wrapper
     * @return hash set containing the points
     */
	public static HashSet <Point> wrapper(Point[] locations) {
		HashSet <Point> wrapper = new HashSet <Point> ();
		Collections.addAll(wrapper, locations);
		return wrapper;
	}
	
    /**
     * Constructs a set of points with locations rotated about 
     * a pivot point.
     * @param locations     locations to be turned
     * @param clockwise     if true rotates clockwise, 
     * counterclockwise otherwise
     * @param pivot     central point about which to turn
     * @return set with rotated points
     */
//  Equation (D = delta (location - pivot)):
//  Clockwise               Counterclockwise
//  x = x - D_x - D_y       x = x - D_x + D_y
//  y = y + D_x - D_y       y = y - D_x - D_y
	public static ArrayList <Point> rotate(Collection <Point> locations, boolean clockwise, Point pivot) {
    	ArrayList <Point> rotated = new ArrayList <Point> ();
		for (Point loc : locations) {
			int x, y;
			if (clockwise) {
				x = pivot.x - loc.y + pivot.y;
				y = pivot.y + loc.x - pivot.x;
			} else {
				x = pivot.x + loc.y - pivot.y;
				y = pivot.y - loc.x + pivot.x;
			}
			rotated.add(new Point(x, y));
		}
		return rotated;
	}
	
    /**
     * Constructs a set of points with a mirror reversal about 
     * both axises along the pivot point. This is the same as 
     * performing two clockwise or counter-clockwise rotations 
     * or both a horizontal and vertical flip.
     * @param locations     locations to be reversed
     * @param pivot     central point about which to reverse
     * @return set with reversed points
     */
	public static HashSet <Point> reverse(Set <Point> locations, Point pivot) {
    	HashSet <Point> reversed = new HashSet <Point> ();
		for (Point loc : locations) {
			int x = loc.x + (pivot.x - loc.x) * 2;
			int y = loc.y + (pivot.y - loc.y) * 2;
			reversed.add(new Point(x, y));
		}
		return reversed;
	}
	
    /**
     * Constructs a set of points with locations shifted to have 
     * a specific upper left location.
     * @param locations     locations to be altered
     * @param target     location of anchor to which points conform
     * @return set with shifted points
     */
	public static HashSet <Point> shift(Set <Point> locations, Point target) {
		Point anchor = getAnchor(locations);
		int xDiff = target.x - anchor.x;
		int yDiff = target.y - anchor.y;
		HashSet <Point> shifted = new HashSet <Point> ();
		for (Point loc : locations) {
			int x = loc.x;
			int y = loc.y;
			shifted.add(new Point(x + xDiff, y + yDiff));
		}
		return shifted;
	}
	
    /**
     * Constructs the shortest path from one point to another. The 
     * collection has the same order as the path. Routes include 
     * two portions: diagnals and streight (if diagnals are included) 
     * and horizontal and vertical (if not). The primary route 
     * places the diagnal or horizontal portions first, the secondary 
     * places them last.
     * @param p1     start of path
     * @param p2     end of path
     * @param diagnals     if true path may include diagnals, 
     * otherwise only strait routes are used
     * @param route     if true the primary path is listed, otherwise 
     * the secondary path is used
     * @return shortest path from one point to another
     */
	public static ArrayList <Point> getPath(Point p1, Point p2, 
			boolean diagnals, boolean route) {
		ArrayList <Point> path = new ArrayList <Point> ();
		path.add(p1);
		while (!p1.equals(p2)) {
			p1 = nextPt(p1, p2, diagnals, route);
			path.add(p1);
		}
		return path;
	}
	
	// Provides the next point in the path between two locations
	// Note: if p1 == p2, this may form an infinite loop
	private static Point nextPt(Point p1, Point p2, boolean diagnals, 
			boolean route) {
//		if (p1.equals(p2)) return null;
		if (diagnals && route) {
			if (p1.x > p2.x && p1.y > p2.y) {
				return new Point(p1.x - 1, p1.y - 1);
			} else if (p1.x < p2.x && p1.y > p2.y) {
				return new Point(p1.x + 1, p1.y - 1);
			} else if (p1.x > p2.x && p1.y < p2.y) {
				return new Point(p1.x - 1, p1.y + 1);
			} else if (p1.x < p2.x && p1.y < p2.y) {
				return new Point(p1.x + 1, p1.y + 1);
			} else {
				return nextPt(p1, p2, true, false);
			}
		} else if (diagnals) {
			int diff = Math.abs(p1.x - p2.x) - Math.abs(p1.y - p2.y);
			if (diff == 0) return nextPt(p1, p2, true, true);
			else if (diff > 0) return nextPt(p1, p2, false, true);
			else return nextPt(p1, p2, false, false);
		} else if (route) {
			if (p1.x > p2.x) return new Point(p1.x - 1, p1.y);
			else if (p1.x < p2.x) return new Point(p1.x + 1, p1.y);
			else return nextPt(p1, p2, false, false);
		} else {
			if (p1.y > p2.y) return new Point(p1.x, p1.y - 1);
			else if (p1.y < p2.y) return new Point(p1.x, p1.y + 1);
			else return nextPt(p1, p2, false, true);
		}
	}
}
